import scanpy as sc
from sklearn import metrics
from scButterfly.logger import *


def calculate_cluster_index(adata):
    """
    Evaluation of cluster index of prediction. 
    Cluster method using Leiden with a default resolution.
    5 different metrics provided: ARI - Adjusted Rand Index, NMI - Normalized Mutual Information, AMI - Adjusted Mutual Information, HOM - Homogeneity, COM - Completeness.
    
    Parameters
    ----------
    adata: Anndata
        Anndata need to cumpute index, there need ground truth labels "cell_type" in adata.obs.
   
    Returns
    ----------
    ARI: float
        Adjusted Rand Index.
    AMI: float
        Normalized Mutual Information.
    NMI: float
        Adjusted Mutual Information.
    HOM: float
        Homogeneity.
    COM: float
        Completeness.
    """
    my_logger = create_logger(name='measure performance', ch=True, fh=False, levelname=logging.INFO, overwrite=False)
    if not "neighbors" in adata.uns.keys():
        my_logger.warning('No "neighbors" in .uns, create "neighbors" first.')
        sc.pp.pca(adata)
        sc.pp.neighbors(adata)
    
    sc.tl.leiden(adata)
    #sc.tl.louvain(adata)
    
    ARI = metrics.adjusted_rand_score(adata.obs['cell_type'], adata.obs['leiden'])
    AMI = metrics.adjusted_mutual_info_score(adata.obs['cell_type'], adata.obs['leiden'])
    NMI = metrics.normalized_mutual_info_score(adata.obs['cell_type'], adata.obs['leiden'])
    HOM = metrics.homogeneity_score(adata.obs['cell_type'], adata.obs['leiden'])
    COM = metrics.completeness_score(adata.obs['cell_type'], adata.obs['leiden'])
    
    return ARI, AMI, NMI, HOM, COM


import scanpy as sc
import anndata as ad
import pandas as pd
import torch
import numpy as np
import random


def setup_seed(seed):
    np.random.seed(seed)
    random.seed(seed)
    torch.cuda.manual_seed_all(seed)
    torch.manual_seed(seed)
    

def idx2adata_multiVI(
    adata,
    train_id,
    validation_id, 
    test_id
):

    train_adata = ad.AnnData(adata.X[train_id, :])
    test_adata = ad.AnnData(adata.X[test_id, :])

    train_adata.var = adata.var
    test_adata.var = adata.var

    train_adata.obs = adata.obs.iloc[train_id, :].copy()
    test_adata.obs = adata.obs.iloc[test_id, :].copy()

    train_adata.obs.reset_index(drop=True, inplace=True)
    test_adata.obs.reset_index(drop=True, inplace=True)

    return train_adata, test_adata



def five_hold_split_dataset(
    RNA_data, 
    ATAC_data, 
    seed = 19193
):
    
    if not seed is None:
        setup_seed(seed)
    
    temp = [i for i in range(len(RNA_data.obs_names))]
    random.shuffle(temp)
    
    id_list = []
    
    test_count = int(0.2 * len(temp))
    validation_count = int(0.16 * len(temp))
    
    for i in range(5):
        test_id = temp[: test_count]
        validation_id = temp[test_count: test_count + validation_count]
        train_id = temp[test_count + validation_count:]
        temp.extend(test_id)
        temp = temp[test_count: ]

        id_list.append([train_id, validation_id, test_id])
    
    return id_list


def celltype_split_dataset(
    RNA_data, 
    ATAC_data, 
    seed = 19193
):
    
    if not seed is None:
        setup_seed(seed)
    
    cell_type_list = list(RNA_data.obs.cell_type)
    cell_type = list(RNA_data.obs.cell_type.cat.categories)

    random.shuffle(cell_type)
    
    id_list = []
    
    test_count = int(len(cell_type) / 3)
    train_count = int(len(cell_type) / 2)
    validation_count = len(cell_type) - test_count - train_count
    
    for i in range(3):
        test_type = cell_type[: test_count]
        validation_type = cell_type[test_count: test_count + validation_count]
        train_type = cell_type[test_count + validation_count:]
        cell_type.extend(test_type)
        cell_type = cell_type[test_count: ]

        train_id = [i for i in range(len(cell_type_list)) if cell_type_list[i] in train_type]
        validation_id = [i for i in range(len(cell_type_list)) if cell_type_list[i] in validation_type]
        test_id = [i for i in range(len(cell_type_list)) if cell_type_list[i] in test_type]

        id_list.append([train_id, validation_id, test_id])
    
    return id_list


def bmmc_batch_split_dataset(
    RNA_data, 
    ATAC_data, 
    seed = 19193
):
    
    if not seed is None:
        setup_seed(seed)
    
    cell_type_list = list(RNA_data.obs_names)
    cell_type = ['1', '2', '3', '4']

    random.shuffle(cell_type)
    
    id_list = []
    
    test_count = 1
    train_count = 2
    validation_count = 1
    
    for i in range(4):
        test_type = cell_type[: test_count]
        validation_type = cell_type[test_count: test_count + validation_count]
        train_type = cell_type[test_count + validation_count:]
        cell_type.extend(test_type)
        cell_type = cell_type[test_count: ]

        train_id = [i for i in range(len(cell_type_list)) if cell_type_list[i].split('-')[-1][1] in train_type]
        validation_id = [i for i in range(len(cell_type_list)) if cell_type_list[i].split('-')[-1][1] in validation_type]
        test_id = [i for i in range(len(cell_type_list)) if cell_type_list[i].split('-')[-1][1] in test_type]

        id_list.append([train_id, validation_id, test_id])
    
    return id_list


def batch_split_dataset(
    RNA_data, 
    ATAC_data, 
    seed = 19193
):
    
    if not seed is None:
        setup_seed(seed)
    
    cell_type_list = list(RNA_data.obs.batch)
    cell_type = list(RNA_data.obs.batch.cat.categories)

    random.shuffle(cell_type)
    
    id_list = []
    
    test_count = int(len(cell_type) / 4)
    validation_count = int(len(cell_type) / 4)
    train_count = len(cell_type) - test_count - validation_count
    
    for i in range(4):
        test_type = cell_type[: test_count]
        validation_type = cell_type[test_count: test_count + validation_count]
        train_type = cell_type[test_count + validation_count:]
        cell_type.extend(test_type)
        cell_type = cell_type[test_count: ]

        train_id = [i for i in range(len(cell_type_list)) if cell_type_list[i] in train_type]
        validation_id = [i for i in range(len(cell_type_list)) if cell_type_list[i] in validation_type]
        test_id = [i for i in range(len(cell_type_list)) if cell_type_list[i] in test_type]

        id_list.append([train_id, validation_id, test_id])
    
    return id_list


def unpaired_split_dataset_yao(
    RNA_data, 
    ATAC_data, 
    seed = 19193
):
    """
    Split datasets into train, validation and test part using different cell types.
    
    Parameters
    ----------
    RNA_data
        full RNA data for spliting.
        
    ATAC_data
        full ATAC data for spliting.
        
    seed
        random seed use to split datasets, if don't give random seed, set it None.
        
    """ 
    
    if not seed is None:
        setup_seed(seed)

    cell_type_list = list(RNA_data.obs.cell_type.cat.categories)
    
    id_list = [[[] for j in range(6)] for i in range(4)]
    
    RNA_batch_list = list(RNA_data.obs.Donor.cat.categories)
    ATAC_batch_list = list(ATAC_data.obs.batch.cat.categories)
    
    random.shuffle(RNA_batch_list)
    random.shuffle(ATAC_batch_list)
    
    max_train_count = 500
    max_validation_count = 200
    
    for i in range(4):
        test_batch_r = RNA_batch_list[0:1]
        validation_batch_r = RNA_batch_list[1:2]
        train_batch_r = RNA_batch_list[2:4]
        test_batch_a = ATAC_batch_list[0:2]
        validation_batch_a = ATAC_batch_list[2:4]
        train_batch_a = ATAC_batch_list[4:9]
        RNA_batch_list.extend(test_batch_r)
        ATAC_batch_list.extend(test_batch_a)
        RNA_batch_list = RNA_batch_list[1:]
        ATAC_batch_list = ATAC_batch_list[2:]
        
        for each in test_batch_r:
            id_list[i][4].extend(list(RNA_data.obs.Donor[RNA_data.obs.Donor == each].index.astype(int)))
            
        for each in test_batch_a:
            id_list[i][5].extend(list(ATAC_data.obs.batch[ATAC_data.obs.batch == each].index.astype(int)))
        
        train_RNA_id = []
        train_ATAC_id = []
        validation_RNA_id = []
        validation_ATAC_id = []
        for each in train_batch_r:
            train_RNA_id.extend(list(RNA_data.obs[RNA_data.obs.Donor == each].index))
        for each in train_batch_a:
            train_ATAC_id.extend(list(ATAC_data.obs[ATAC_data.obs.batch == each].index))
        for each in validation_batch_r:
            validation_RNA_id.extend(list(RNA_data.obs[RNA_data.obs.Donor == each].index))
        for each in validation_batch_a:
            validation_ATAC_id.extend(list(ATAC_data.obs[ATAC_data.obs.batch == each].index))
        train_RNA = RNA_data.obs.loc[train_RNA_id, :]
        train_ATAC = ATAC_data.obs.loc[train_ATAC_id, :]
        validation_RNA = RNA_data.obs.loc[validation_RNA_id, :]
        validation_ATAC = ATAC_data.obs.loc[validation_ATAC_id, :]
        
        for ctype in cell_type_list:
            id_temp = list(train_RNA[train_RNA.cell_type == ctype].index.astype(int))
            id_list[i][0].extend(np.random.choice(id_temp, max_train_count, replace=True))     
            id_temp = list(train_ATAC[train_ATAC.cell_type == ctype].index.astype(int))
            id_list[i][1].extend(np.random.choice(id_temp, max_train_count, replace=True))
            id_temp = list(validation_RNA[validation_RNA.cell_type == ctype].index.astype(int))
            id_list[i][2].extend(np.random.choice(id_temp, max_validation_count, replace=True))
            id_temp = list(validation_ATAC[validation_ATAC.cell_type == ctype].index.astype(int))
            id_list[i][3].extend(np.random.choice(id_temp, max_validation_count, replace=True))
        
    return id_list


def unpaired_split_dataset_muto(
    RNA_data, 
    ATAC_data, 
    seed = 19193
):
    """
    Split datasets into train, validation and test part using different cell types.
    
    Parameters
    ----------
    RNA_data
        full RNA data for spliting.
        
    ATAC_data
        full ATAC data for spliting.
        
    seed
        random seed use to split datasets, if don't give random seed, set it None.
        
    """ 
    
    if not seed is None:
        setup_seed(seed)

    cell_type_list = list(RNA_data.obs.cell_type.cat.categories)
    
    id_list = [[[] for j in range(6)] for i in range(5)]
    
    batch_list = list(RNA_data.obs.batch.cat.categories)
    
    random.shuffle(batch_list)
    
    max_train_count = 500
    max_validation_count = 200
    
    for i in range(5):
        test_batch_r = batch_list[0:1]
        validation_batch_r = batch_list[1:2]
        train_batch_r = batch_list[2:5]
        test_batch_a = batch_list[0:1]
        validation_batch_a = batch_list[1:2]
        train_batch_a = batch_list[2:5]
        batch_list.extend(test_batch_r)
        batch_list = batch_list[1:]
        
        for each in test_batch_r:
            id_list[i][4].extend(list(RNA_data.obs.batch[RNA_data.obs.batch == each].index.astype(int)))
            
        for each in test_batch_a:
            id_list[i][5].extend(list(ATAC_data.obs.batch[ATAC_data.obs.batch == each].index.astype(int)))
        
        train_RNA_id = []
        train_ATAC_id = []
        validation_RNA_id = []
        validation_ATAC_id = []
        for each in train_batch_r:
            train_RNA_id.extend(list(RNA_data.obs[RNA_data.obs.batch == each].index))
        for each in train_batch_a:
            train_ATAC_id.extend(list(ATAC_data.obs[ATAC_data.obs.batch == each].index))
        for each in validation_batch_r:
            validation_RNA_id.extend(list(RNA_data.obs[RNA_data.obs.batch == each].index))
        for each in validation_batch_a:
            validation_ATAC_id.extend(list(ATAC_data.obs[ATAC_data.obs.batch == each].index))
        train_RNA = RNA_data.obs.loc[train_RNA_id, :]
        train_ATAC = ATAC_data.obs.loc[train_ATAC_id, :]
        validation_RNA = RNA_data.obs.loc[validation_RNA_id, :]
        validation_ATAC = ATAC_data.obs.loc[validation_ATAC_id, :]
        
        for ctype in cell_type_list:
            id_temp = list(train_RNA[train_RNA.cell_type == ctype].index.astype(int))
            id_list[i][0].extend(np.random.choice(id_temp, max_train_count, replace=True))     
            id_temp = list(train_ATAC[train_ATAC.cell_type == ctype].index.astype(int))
            id_list[i][1].extend(np.random.choice(id_temp, max_train_count, replace=True))
            id_temp = list(validation_RNA[validation_RNA.cell_type == ctype].index.astype(int))
            id_list[i][2].extend(np.random.choice(id_temp, max_validation_count, replace=True))
            id_temp = list(validation_ATAC[validation_ATAC.cell_type == ctype].index.astype(int))
            id_list[i][3].extend(np.random.choice(id_temp, max_validation_count, replace=True))
        
    return id_list


def unpaired_split_dataset_yao_scglue(
    RNA_data, 
    ATAC_data, 
    combine_list,
    seed = 19193
):
    """
    Split datasets into train, validation and test part using different cell types.
    
    Parameters
    ----------
    RNA_data
        full RNA data for spliting.
        
    ATAC_data
        full ATAC data for spliting.
        
    seed
        random seed use to split datasets, if don't give random seed, set it None.
        
    """ 
    
    if not seed is None:
        setup_seed(seed)
    
    id_list = [[[] for j in range(6)] for i in range(4)]
    
    RNA_batch_list = list(RNA_data.obs.Donor.cat.categories)
    ATAC_batch_list = list(ATAC_data.obs.batch.cat.categories)
    
    random.shuffle(RNA_batch_list)
    random.shuffle(ATAC_batch_list)
    
    target_train = len(list(RNA_data.obs.cell_type.cat.categories)) * 500
    target_validation = len(list(RNA_data.obs.cell_type.cat.categories)) * 200
    
    for i in range(4):
        combine = combine_list[i]
        cell_type_list = list(combine.obs.leiden.cat.categories)
        combine_RNA = combine[combine.obs.domain == 'scRNA-seq']
        combine_ATAC = combine[combine.obs.domain == 'scATAC-seq']
        
        max_train_count = int(target_train / len(cell_type_list)) + 1
        max_validation_count = int(target_validation / len(cell_type_list)) + 1
        
        test_batch_r = RNA_batch_list[0:1]
        validation_batch_r = RNA_batch_list[1:2]
        train_batch_r = RNA_batch_list[2:4]
        test_batch_a = ATAC_batch_list[0:2]
        validation_batch_a = ATAC_batch_list[2:4]
        train_batch_a = ATAC_batch_list[4:9]
        RNA_batch_list.extend(test_batch_r)
        ATAC_batch_list.extend(test_batch_a)
        RNA_batch_list = RNA_batch_list[1:]
        ATAC_batch_list = ATAC_batch_list[2:]
        
        for each in test_batch_r:
            id_list[i][4].extend(list(RNA_data.obs.Donor[RNA_data.obs.Donor == each].index.astype(int)))
            
        for each in test_batch_a:
            id_list[i][5].extend(list(ATAC_data.obs.batch[ATAC_data.obs.batch == each].index.astype(int)))
        
        train_RNA_id = []
        train_ATAC_id = []
        validation_RNA_id = []
        validation_ATAC_id = []
        for each in train_batch_r:
            train_RNA_id.extend(list(RNA_data.obs[RNA_data.obs.Donor == each].index))
        for each in train_batch_a:
            train_ATAC_id.extend(list(ATAC_data.obs[ATAC_data.obs.batch == each].index))
        for each in validation_batch_r:
            validation_RNA_id.extend(list(RNA_data.obs[RNA_data.obs.Donor == each].index))
        for each in validation_batch_a:
            validation_ATAC_id.extend(list(ATAC_data.obs[ATAC_data.obs.batch == each].index))
        train_RNA = combine_RNA.obs.loc[train_RNA_id, :]
        train_ATAC = combine_ATAC.obs.loc[train_ATAC_id, :]
        validation_RNA = combine_RNA.obs.loc[validation_RNA_id, :]
        validation_ATAC = combine_ATAC.obs.loc[validation_ATAC_id, :]
        for ctype in cell_type_list:
            id_temp_train_RNA = list(train_RNA[train_RNA.leiden == ctype].index.astype(int))    
            id_temp_train_ATAC = list(train_ATAC[train_ATAC.leiden == ctype].index.astype(int))
            id_temp_validation_RNA = list(validation_RNA[validation_RNA.leiden == ctype].index.astype(int))
            id_temp_validation_ATAC = list(validation_ATAC[validation_ATAC.leiden == ctype].index.astype(int))
            if (len(id_temp_train_RNA) != 0) and (len(id_temp_train_ATAC) != 0):
                id_list[i][0].extend(np.random.choice(id_temp_train_RNA, max_train_count, replace=True))
                id_list[i][1].extend(np.random.choice(id_temp_train_ATAC, max_train_count, replace=True))
            if (len(id_temp_validation_RNA) != 0) and (len(id_temp_validation_ATAC) != 0):    
                id_list[i][2].extend(np.random.choice(id_temp_validation_RNA, max_validation_count, replace=True))
                id_list[i][3].extend(np.random.choice(id_temp_validation_ATAC, max_validation_count, replace=True))

    return id_list


def unpaired_split_dataset_muto_scglue(
    RNA_data, 
    ATAC_data, 
    combine_list,
    seed = 19193
):
    """
    Split datasets into train, validation and test part using different cell types.
    
    Parameters
    ----------
    RNA_data
        full RNA data for spliting.
        
    ATAC_data
        full ATAC data for spliting.
        
    seed
        random seed use to split datasets, if don't give random seed, set it None.
        
    """ 
    
    if not seed is None:
        setup_seed(seed)
    
    id_list = [[[] for j in range(6)] for i in range(5)]
    
    batch_list = list(RNA_data.obs.batch.cat.categories)
    
    random.shuffle(batch_list)
    
    target_train = len(list(RNA_data.obs.cell_type.cat.categories)) * 500
    target_validation = len(list(RNA_data.obs.cell_type.cat.categories)) * 200
    
    for i in range(5):
        combine = combine_list[i]
        cell_type_list = list(combine.obs.leiden.cat.categories)
        combine_RNA = combine[combine.obs.domain == 'scRNA-seq']
        combine_ATAC = combine[combine.obs.domain == 'scATAC-seq']
        
        max_train_count = int(target_train / len(cell_type_list)) + 1
        max_validation_count = int(target_validation / len(cell_type_list)) + 1

        test_batch_r = batch_list[0:1]
        validation_batch_r = batch_list[1:2]
        train_batch_r = batch_list[2:5]
        test_batch_a = batch_list[0:1]
        validation_batch_a = batch_list[1:2]
        train_batch_a = batch_list[2:5]
        batch_list.extend(test_batch_r)
        batch_list = batch_list[1:]
        
        for each in test_batch_r:
            id_list[i][4].extend(list(RNA_data.obs.batch[RNA_data.obs.batch == each].index.astype(int)))
            
        for each in test_batch_a:
            id_list[i][5].extend(list(ATAC_data.obs.batch[ATAC_data.obs.batch == each].index.astype(int)))
        
        train_RNA_id = []
        train_ATAC_id = []
        validation_RNA_id = []
        validation_ATAC_id = []
        for each in train_batch_r:
            train_RNA_id.extend(list(RNA_data.obs[RNA_data.obs.batch == each].index))
        for each in train_batch_a:
            train_ATAC_id.extend(list(ATAC_data.obs[ATAC_data.obs.batch == each].index))
        for each in validation_batch_r:
            validation_RNA_id.extend(list(RNA_data.obs[RNA_data.obs.batch == each].index))
        for each in validation_batch_a:
            validation_ATAC_id.extend(list(ATAC_data.obs[ATAC_data.obs.batch == each].index))
        train_RNA = combine_RNA.obs.loc[train_RNA_id, :]
        train_ATAC = combine_ATAC.obs.loc[train_ATAC_id, :]
        validation_RNA = combine_RNA.obs.loc[validation_RNA_id, :]
        validation_ATAC = combine_ATAC.obs.loc[validation_ATAC_id, :]
        for ctype in cell_type_list:
            id_temp_train_RNA = list(train_RNA[train_RNA.leiden == ctype].index.astype(int))    
            id_temp_train_ATAC = list(train_ATAC[train_ATAC.leiden == ctype].index.astype(int))
            id_temp_validation_RNA = list(validation_RNA[validation_RNA.leiden == ctype].index.astype(int))
            id_temp_validation_ATAC = list(validation_ATAC[validation_ATAC.leiden == ctype].index.astype(int))
            if (len(id_temp_train_RNA) != 0) and (len(id_temp_train_ATAC) != 0):
                id_list[i][0].extend(np.random.choice(id_temp_train_RNA, max_train_count, replace=True))
                id_list[i][1].extend(np.random.choice(id_temp_train_ATAC, max_train_count, replace=True))
            if (len(id_temp_validation_RNA) != 0) and (len(id_temp_validation_ATAC) != 0):    
                id_list[i][2].extend(np.random.choice(id_temp_validation_RNA, max_validation_count, replace=True))
                id_list[i][3].extend(np.random.choice(id_temp_validation_ATAC, max_validation_count, replace=True))
        
    return id_list


def seurat_batch_split_dataset(
    RNA_data, 
    ATAC_data, 
    seed = 19193
):
    """
    Split datasets into train, validation and test part using different cell types.
    
    Parameters
    ----------
    RNA_data
        full RNA data for spliting.
        
    ATAC_data
        full ATAC data for spliting.
        
    seed
        random seed use to split datasets, if don't give random seed, set it None.
        
    """ 
    
    if not seed is None:
        setup_seed(seed)
    
    cell_type_list = list(RNA_data.obs.batch)
    cell_type = list(RNA_data.obs.batch.cat.categories)

    random.shuffle(cell_type)
    
    id_list = []
    
    for i in range(2):
        test_type = cell_type[i]
        train_type = cell_type[i - 1]
        
        train_valid_id = [i for i in range(len(cell_type_list)) if cell_type_list[i] == train_type]
        test_id = [i for i in range(len(cell_type_list)) if cell_type_list[i] == test_type]

        random.shuffle(train_valid_id)
        
        train_id = train_valid_id[0: int(0.8 * len(train_valid_id))]
        validation_id = train_valid_id[int(0.8 * len(train_valid_id)):]
        
        id_list.append([train_id, validation_id, test_id])
    
    return id_list


def unpaired_split_dataset_perturb(
    RNA_data, 
    ATAC_data, 
    seed = 19193
):
    """
    Split datasets into train, validation and test part using different cell types.
    
    Parameters
    ----------
    RNA_data
        full RNA data for spliting.
        
    ATAC_data
        full ATAC data for spliting.
        
    seed
        random seed use to split datasets, if don't give random seed, set it None.
        
    """ 
    
    if not seed is None:
        setup_seed(seed)

    cell_type_list = list(RNA_data.obs.cell_type.cat.categories)
    
    id_list = [[[] for j in range(6)] for i in range(len(cell_type_list))]
    
    batch_list = list(RNA_data.obs.cell_type.cat.categories)
    
    random.shuffle(batch_list)
    
    max_train_count = 1200
    max_validation_count = 800
    
    for i in range(len(cell_type_list)):
        
        test_batch_r = batch_list[:1]
        validation_batch_r = batch_list[1:]
        train_batch_r = batch_list[1:]
        test_batch_a = batch_list[:1]
        validation_batch_a = batch_list[1:]
        train_batch_a = batch_list[1:]
        batch_list.extend(test_batch_r)
        batch_list = batch_list[1:]
        
        for each in test_batch_r:
            id_list[i][4].extend(list(RNA_data.obs.cell_type[RNA_data.obs.cell_type == each].index.astype(int)))
            
        for each in test_batch_a:
            id_list[i][5].extend(list(ATAC_data.obs.cell_type[ATAC_data.obs.cell_type == each].index.astype(int)))
        
        train_RNA_id = []
        train_ATAC_id = []
        validation_RNA_id = []
        validation_ATAC_id = []
        for each in train_batch_r:
            train_RNA_id.extend(list(RNA_data.obs[RNA_data.obs.cell_type == each].index))
        for each in train_batch_a:
            train_ATAC_id.extend(list(ATAC_data.obs[ATAC_data.obs.cell_type == each].index))
        for each in validation_batch_r:
            validation_RNA_id.extend(list(RNA_data.obs[RNA_data.obs.cell_type == each].index))
        for each in validation_batch_a:
            validation_ATAC_id.extend(list(ATAC_data.obs[ATAC_data.obs.cell_type == each].index))
        train_RNA = RNA_data.obs.loc[train_RNA_id, :]
        train_ATAC = ATAC_data.obs.loc[train_ATAC_id, :]
        validation_RNA = RNA_data.obs.loc[validation_RNA_id, :]
        validation_ATAC = ATAC_data.obs.loc[validation_ATAC_id, :]
        
        for ctype in cell_type_list:
            max_train_count_temp = int(train_RNA.cell_type.value_counts(normalize=True)[ctype] * max_train_count)
            max_validation_count_temp = int(validation_RNA.cell_type.value_counts(normalize=True)[ctype] * max_validation_count)
            
            id_temp = list(train_RNA[train_RNA.cell_type == ctype].index.astype(int))
            if len(id_temp) != 0:
                id_list[i][0].extend(np.random.choice(id_temp, max_train_count_temp, replace=True))     
            id_temp = list(train_ATAC[train_ATAC.cell_type == ctype].index.astype(int))
            if len(id_temp) != 0:
                id_list[i][1].extend(np.random.choice(id_temp, max_train_count_temp, replace=True))
            id_temp = list(validation_RNA[validation_RNA.cell_type == ctype].index.astype(int))
            if len(id_temp) != 0:
                id_list[i][2].extend(np.random.choice(id_temp, max_validation_count_temp, replace=True))
            id_temp = list(validation_ATAC[validation_ATAC.cell_type == ctype].index.astype(int))
            if len(id_temp) != 0:
                id_list[i][3].extend(np.random.choice(id_temp, max_validation_count_temp, replace=True))
        
    return id_list


import imp
from statistics import mean
from turtle import forward
from matplotlib.pyplot import cla
import numpy as np
import pandas as pd
import torch
import torch.nn as nn
import torch.nn.functional as F
from sklearn.metrics import roc_auc_score, precision_recall_curve, auc
from scipy.stats import spearmanr, pearsonr
import sys
from scipy import integrate

from urllib3 import Retry


#binrize ATAC
def binirize(atac_seq, every_mean):
    bin = []
    
    for i in range (len(atac_seq)):
        if atac_seq[i] < every_mean[i] :
            bin.append(0)
            #bin.append(atac_seq[i])
            
        else:
            bin.append(1)
            #bin.append(atac_seq[i])
    bin = np.array(bin)
    return bin

#PP expression rate
def PP (atac_seq):
    sum = 0
    for i in range(len(atac_seq)):
        if atac_seq[i]>0:
            sum += 1
    return sum/len(atac_seq)

#AUROC: RNA to ATAC
def AUROC(atac_seq_pre,atac_seq, every_mean):
    auroc = 0
    atac_seq = binirize(atac_seq,every_mean)

    try:
        auroc = roc_auc_score(atac_seq, atac_seq_pre)
    except ValueError:
        pass

    #auroc = roc_auc_score(atac_seq,atac_seq_pre)
    return auroc
def draw_AUROC(atac_seq_pre,atac_seq, every_mean,output_path):
    auroc = 0
    atac_seq = binirize(atac_seq,every_mean)
    fpr, tpr, thresholds = roc_curve(atac_seq, atac_seq_pre)
    pyplot.plot([0, 1], [0, 1], linestyle='--')
    pyplot.plot(fpr, tpr, marker='.')
    pyplot.title("ROC")
    pyplot.xlabel("FPR")
    pyplot.ylabel("TPR")
    pyplot.savefig(output_path+'AUROC.pdf')
    #auroc = roc_auc_score(atac_seq,atac_seq_pre)
    return auroc

#AUPRC: RNA to ATAC
def AUPR_norm (atac_seq_pre,atac_seq, every_mean):
    atac_seq = binirize(atac_seq, every_mean)
    norm = 0  
    presion, recall, thee = precision_recall_curve(atac_seq,atac_seq_pre)
    sorted_index = np.argsort(presion)
    fpr_list_sorted =  np.array(presion)[sorted_index]
    tpr_list_sorted = np.array(recall)[sorted_index]
    integrate.trapz(y=tpr_list_sorted, x=fpr_list_sorted)
    aup = auc(fpr_list_sorted, tpr_list_sorted)
    pp = PP(atac_seq)
    norm = (aup-pp)/(1-pp)
    return aup

def draw_AUPR_norm (atac_seq_pre,atac_seq, every_mean,output_path):
    atac_seq = binirize(atac_seq, every_mean)
    norm = 0  
    presion, recall, thee = precision_recall_curve(atac_seq,atac_seq_pre)
    pyplot.plot(reacall, precision, 'r')
    pyplot.legend(loc='lower left')
    pyplot.xlabel('Recall')  # the name of x-axis
    pyplot.ylabel('Precision')  # the name of y-axis
    pyplot.title('Precision-Recall')  # the title of figure
    pyplot.savefig(output_path+'presicion_recall.pdf')
    return aup
#Pearson ATAC to RNA
def pearson (rna_seq_pre,rna_seq):
    person = pearsonr(rna_seq,rna_seq_pre)[0]
    return person

#Spearman ATAC to RNA
def spearman (rna_seq_pre,rna_seq):
    spe = spearmanr(rna_seq,rna_seq_pre)[0]
    return spe

#FOSCTTM 
def foscttm (atac_seq_pre,total_samples):
    fracs=calc_domainAveraged_FOSCTTM(total_samples, atac_seq_pre)
    fracs = np.mean(fracs)
    return fracs

#Only need this function to test a model
def evaluate(My_model, test_dataloader, batch_size, RNA_input_dim, ATAC_input_dim,loss_func, every_mean):
    
    mean_test_loss = 0
    mean_test_auroc = 0
    mean_test_aupr = 0
    mean_test_pearson = 0
    mean_test_spearman = 0
    sum_auroc = 0
    sum_aupr = 0
    max_enum = 9
    for idx, batch_samples in enumerate(test_dataloader):
        if torch.cuda.is_available():
            batch_samples = batch_samples.cuda()
        

        RNA_input, ATAC_input = torch.split(batch_samples, [RNA_input_dim, ATAC_input_dim], dim=1)
        R2R, R2A = My_model(RNA_input,'RNA')
        A2R, A2A = My_model(ATAC_input,'ATAC')
        R2, A2 = torch.cat([R2R, R2A], dim=1), torch.cat([A2R, A2A], dim=1)
        lossR_temp = loss_func(R2, batch_samples)
        lossA_temp = loss_func(A2, batch_samples)
        loss_temp = lossA_temp + lossR_temp


        atac_input = ATAC_input.cpu().numpy()
        rna_input = RNA_input.cpu().numpy()
        r2a = R2A.cpu().detach().numpy()
        a2r = A2R.cpu().detach().numpy()
        ss = sum(atac_input)

        #对每一个cell， 计算auroc和aupr，再取平均
        for i in range(len(atac_input)):
            mean_test_loss += loss_temp.item()
            a=AUROC(r2a[i,:],atac_input[i,:], every_mean)
            if a != 0:
                mean_test_auroc += a
                sum_auroc += 1
            else:
                mean_test_auroc += a
            b = AUPR_norm(r2a[i,:],atac_input[i,:], every_mean)
            if b!= 0:
                mean_test_aupr += b
                sum_aupr += 1
            else:
                mean_test_aupr += b
            
            mean_test_pearson += pearson(a2r[i,:],rna_input[i,:])
            mean_test_spearman += spearman(a2r[i,:],rna_input[i,:])

            finish = idx/len(test_dataloader)
            print(finish,"has been finished.")

        

    mean_test_loss /= len(test_dataloader)*batch_size
    mean_test_auroc /= sum_auroc
    mean_test_aupr /= sum_aupr
    mean_test_pearson /= len(test_dataloader)*batch_size
    mean_test_spearman /= len(test_dataloader)*batch_size
    print('*****mean_test_loss: ' + str(mean_test_loss) + '*****')
    print('*****mean_test_auroc: ' + str(mean_test_auroc) + '*****')
    print('*****mean_test_aupr: ' + str(mean_test_aupr) + '*****')
    print('*****mean_test_pearson: ' + str(mean_test_pearson) + '*****')
    print('*****mean_test_spearman: ' + str(mean_test_spearman) + '*****')
    return [mean_test_loss,mean_test_auroc,mean_test_aupr,mean_test_pearson,mean_test_spearman]




import torch
import torch.nn as nn
import torch.nn.functional as F


class NetBlock(nn.Module):
    def __init__(
        self,
        nlayer: int, 
        dim_list: list, 
        act_list: list, 
        dropout_rate: float, 
        noise_rate: float
        ):
        """
        multiple layers netblock with specific layer counts, dimension, activations and dropout.
        
        Parameters
        ----------
        nlayer
            layer counts.
            
        dim_list
            dimension list, length equal to nlayer + 1.
        
        act_list
            activation list, length equal to nlayer + 1.
        
        dropout_rate
            rate of dropout.
        
        noise_rate
            rate of set part of input data to 0.
            
        """
        super(NetBlock, self).__init__()
        self.nlayer = nlayer
        self.noise_dropout = nn.Dropout(noise_rate)
        self.linear_list = nn.ModuleList()
        self.bn_list = nn.ModuleList()
        self.activation_list = nn.ModuleList()
        self.dropout_list = nn.ModuleList()
        
        for i in range(nlayer):
            
            self.linear_list.append(nn.Linear(dim_list[i], dim_list[i + 1]))
            nn.init.xavier_uniform_(self.linear_list[i].weight)
            self.bn_list.append(nn.BatchNorm1d(dim_list[i + 1]))
            self.activation_list.append(act_list[i])
            if not i == nlayer -1: 
                self.dropout_list.append(nn.Dropout(dropout_rate))
        
    
    def forward(self, x):

        x = self.noise_dropout(x)
        for i in range(self.nlayer):
            x = self.linear_list[i](x)
            x = self.bn_list[i](x)
            x = self.activation_list[i](x)
            if not i == self.nlayer -1:
                """ don't use dropout for output to avoid loss calculate break down """
                x = self.dropout_list[i](x)

        return x

    
class Split_Chrom_Encoder_block(nn.Module):
    def __init__(
        self,
        nlayer: int, 
        dim_list: list, 
        act_list: list,
        chrom_list: list,
        dropout_rate: float, 
        noise_rate: float
        ):
        """
        ATAC encoder netblock with specific layer counts, dimension, activations and dropout.
        
        Parameters
        ----------
        nlayer
            layer counts.
            
        dim_list
            dimension list, length equal to nlayer + 1.
        
        act_list
            activation list, length equal to nlayer + 1.
            
        chrom_list
            list record the peaks count for each chrom, assert that sum of chrom list equal to dim_list[0].
            
        dropout_rate
            rate of dropout.
        
        noise_rate
            rate of set part of input data to 0.
            
        """
        super(Split_Chrom_Encoder_block, self).__init__()
        self.nlayer = nlayer
        self.chrom_list = chrom_list
        self.noise_dropout = nn.Dropout(noise_rate)
        self.linear_list = nn.ModuleList()
        self.bn_list = nn.ModuleList()
        self.activation_list = nn.ModuleList()
        self.dropout_list = nn.ModuleList()
        
        for i in range(nlayer):
            if i == 0:
                """first layer seperately forward for each chrom"""
                self.linear_list.append(nn.ModuleList())
                self.bn_list.append(nn.ModuleList())
                self.activation_list.append(nn.ModuleList())
                self.dropout_list.append(nn.ModuleList())
                for j in range(len(chrom_list)):
                    self.linear_list[i].append(nn.Linear(chrom_list[j], dim_list[i + 1] // len(chrom_list)))
                    nn.init.xavier_uniform_(self.linear_list[i][j].weight)
                    self.bn_list[i].append(nn.BatchNorm1d(dim_list[i + 1] // len(chrom_list)))
                    self.activation_list[i].append(act_list[i])
                    self.dropout_list[i].append(nn.Dropout(dropout_rate))
            else:
                self.linear_list.append(nn.Linear(dim_list[i], dim_list[i + 1]))
                nn.init.xavier_uniform_(self.linear_list[i].weight)
                self.bn_list.append(nn.BatchNorm1d(dim_list[i + 1]))
                self.activation_list.append(act_list[i])
                if not i == nlayer -1: 
                    self.dropout_list.append(nn.Dropout(dropout_rate))
        
    
    def forward(self, x):

        x = self.noise_dropout(x)
        for i in range(self.nlayer):
            if i == 0:
                x = torch.split(x, self.chrom_list, dim = 1)
                temp = []
                for j in range(len(self.chrom_list)):
                    temp.append(self.dropout_list[0][j](self.activation_list[0][j](self.bn_list[0][j](self.linear_list[0][j](x[j])))))
                x = torch.concat(temp, dim = 1)
            else:
                x = self.linear_list[i](x)
                x = self.bn_list[i](x)
                x = self.activation_list[i](x)
                if not i == self.nlayer -1:
                    """ don't use dropout for output to avoid loss calculate break down """
                    x = self.dropout_list[i](x)
        return x


class Split_Chrom_Decoder_block(nn.Module):
    def __init__(
        self,
        nlayer: int, 
        dim_list: list, 
        act_list: list,
        chrom_list: list,
        dropout_rate: float, 
        noise_rate: float
        ):
        """
        ATAC decoder netblock with specific layer counts, dimension, activations and dropout.
        
        Parameters
        ----------
        nlayer
            layer counts.
            
        dim_list
            dimension list, length equal to nlayer + 1.
        
        act_list
            activation list, length equal to nlayer + 1.
            
        chrom_list
            list record the peaks count for each chrom, assert that sum of chrom list equal to dim_list[end].
            
        dropout_rate
            rate of dropout.
        
        noise_rate
            rate of set part of input data to 0.
            
        """
        super(Split_Chrom_Decoder_block, self).__init__()
        self.nlayer = nlayer
        self.noise_dropout = nn.Dropout(noise_rate)
        self.chrom_list = chrom_list
        self.linear_list = nn.ModuleList()
        self.bn_list = nn.ModuleList()
        self.activation_list = nn.ModuleList()
        self.dropout_list = nn.ModuleList()
        
        for i in range(nlayer):
            if not i == nlayer -1:
                self.linear_list.append(nn.Linear(dim_list[i], dim_list[i + 1]))
                nn.init.xavier_uniform_(self.linear_list[i].weight)
                self.bn_list.append(nn.BatchNorm1d(dim_list[i + 1]))
                self.activation_list.append(act_list[i])
                self.dropout_list.append(nn.Dropout(dropout_rate))
            else:
                """last layer seperately forward for each chrom"""
                self.linear_list.append(nn.ModuleList())
                self.bn_list.append(nn.ModuleList())
                self.activation_list.append(nn.ModuleList())
                self.dropout_list.append(nn.ModuleList())
                for j in range(len(chrom_list)):
                    self.linear_list[i].append(nn.Linear(dim_list[i] // len(chrom_list), chrom_list[j]))
                    nn.init.xavier_uniform_(self.linear_list[i][j].weight)
                    self.bn_list[i].append(nn.BatchNorm1d(chrom_list[j]))
                    self.activation_list[i].append(act_list[i])
        
    
    def forward(self, x):

        x = self.noise_dropout(x)
        for i in range(self.nlayer):
            if not i == self.nlayer -1:
                x = self.linear_list[i](x)
                x = self.bn_list[i](x)
                x = self.activation_list[i](x)
                x = self.dropout_list[i](x)
            else:
                x = torch.chunk(x, len(self.chrom_list), dim = 1)
                temp = []
                for j in range(len(self.chrom_list)):
                    temp.append(self.activation_list[i][j](self.bn_list[i][j](self.linear_list[i][j](x[j]))))
                x = torch.concat(temp, dim = 1)

        return x
    
    
class Translator(nn.Module):
    def __init__(
        self,
        translator_input_dim_r: int, 
        translator_input_dim_a: int, 
        translator_embed_dim: int, 
        translator_embed_act_list: list, 
        ):
        """
        Translator block used for translation between different omics.
        
        Parameters
        ----------
        translator_input_dim_r
            dimension of input from RNA encoder for translator.
            
        translator_input_dim_a
            dimension of input from ATAC encoder for translator.
            
        translator_embed_dim
            dimension of embedding space for translator.
            
        translator_embed_act_list
            activation list for translator, involving [mean_activation, log_var_activation, decoder_activation].
            
        """
        
        super(Translator, self).__init__()
        
        mean_activation, log_var_activation, decoder_activation = translator_embed_act_list
        
        self.RNA_encoder_l_mu = nn.Linear(translator_input_dim_r, translator_embed_dim)
        nn.init.xavier_uniform_(self.RNA_encoder_l_mu.weight)
        self.RNA_encoder_bn_mu = nn.BatchNorm1d(translator_embed_dim)
        self.RNA_encoder_act_mu = mean_activation
        
        self.ATAC_encoder_l_mu = nn.Linear(translator_input_dim_a, translator_embed_dim)
        nn.init.xavier_uniform_(self.ATAC_encoder_l_mu.weight)
        self.ATAC_encoder_bn_mu = nn.BatchNorm1d(translator_embed_dim)
        self.ATAC_encoder_act_mu = mean_activation
        
        self.RNA_encoder_l_d = nn.Linear(translator_input_dim_r, translator_embed_dim)
        nn.init.xavier_uniform_(self.RNA_encoder_l_d.weight)
        self.RNA_encoder_bn_d = nn.BatchNorm1d(translator_embed_dim)
        self.RNA_encoder_act_d = log_var_activation
        
        self.ATAC_encoder_l_d = nn.Linear(translator_input_dim_a, translator_embed_dim)
        nn.init.xavier_uniform_(self.ATAC_encoder_l_d.weight)
        self.ATAC_encoder_bn_d = nn.BatchNorm1d(translator_embed_dim)
        self.ATAC_encoder_act_d = log_var_activation
        
        self.RNA_decoder_l = nn.Linear(translator_embed_dim, translator_input_dim_r)
        nn.init.xavier_uniform_(self.RNA_decoder_l.weight)
        self.RNA_decoder_bn = nn.BatchNorm1d(translator_input_dim_r)
        self.RNA_decoder_act = decoder_activation
        
        self.ATAC_decoder_l = nn.Linear(translator_embed_dim, translator_input_dim_a)
        nn.init.xavier_uniform_(self.ATAC_decoder_l.weight)
        self.ATAC_decoder_bn = nn.BatchNorm1d(translator_input_dim_a)
        self.ATAC_decoder_act = decoder_activation
                
    
    def reparameterize(self, mu, sigma):
        sigma = torch.exp(sigma/2)
        eps = torch.randn_like(sigma)
        return mu + eps * sigma
    
    def forward_with_RNA(self, x, forward_type):
            
        latent_layer_mu = self.RNA_encoder_act_mu(self.RNA_encoder_bn_mu(self.RNA_encoder_l_mu(x)))
        latent_layer_d = self.RNA_encoder_act_d(self.RNA_encoder_bn_d(self.RNA_encoder_l_d(x)))
        
        if forward_type == 'test':
            latent_layer = latent_layer_mu
        elif forward_type == 'train':
            latent_layer = self.reparameterize(latent_layer_mu, latent_layer_d)
        
        latent_layer_R_out = self.RNA_decoder_act(self.RNA_decoder_bn(self.RNA_decoder_l(latent_layer)))
        latent_layer_A_out = self.ATAC_decoder_act(self.ATAC_decoder_bn(self.ATAC_decoder_l(latent_layer)))

        return latent_layer_R_out, latent_layer_A_out, latent_layer_mu, latent_layer_d
    
    def forward_with_ATAC(self, x, forward_type):
        # the function with ATAC input and RNA, ATAC output 
            
        latent_layer_mu = self.ATAC_encoder_act_mu(self.ATAC_encoder_bn_mu(self.ATAC_encoder_l_mu(x)))
        latent_layer_d = self.ATAC_encoder_act_d(self.ATAC_encoder_bn_d(self.ATAC_encoder_l_d(x)))
        
        if forward_type == 'test':
            latent_layer = latent_layer_mu
        elif forward_type == 'train':
            latent_layer = self.reparameterize(latent_layer_mu, latent_layer_d)
        
        latent_layer_R_out = self.RNA_decoder_act(self.RNA_decoder_bn(self.RNA_decoder_l(latent_layer)))
        latent_layer_A_out = self.ATAC_decoder_act(self.ATAC_decoder_bn(self.ATAC_decoder_l(latent_layer)))

        return latent_layer_R_out, latent_layer_A_out, latent_layer_mu, latent_layer_d
    
    def train_model(self, x, input_type):
        if input_type == 'RNA':
            latent_layer_R_out, latent_layer_A_out, latent_layer_mu, latent_layer_d = self.forward_with_RNA(x, 'train')
            return latent_layer_R_out, latent_layer_A_out, latent_layer_mu, latent_layer_d
        elif input_type == 'ATAC':
            latent_layer_R_out, latent_layer_A_out, latent_layer_mu, latent_layer_d = self.forward_with_ATAC(x, 'train')
            return latent_layer_R_out, latent_layer_A_out, latent_layer_mu, latent_layer_d
        
    def test_model(self, x, input_type):
        if input_type == 'RNA':
            latent_layer_R_out, latent_layer_A_out, latent_layer_mu, latent_layer_d = self.forward_with_RNA(x, 'test')
            return latent_layer_R_out, latent_layer_A_out, latent_layer_mu, latent_layer_d
        elif input_type == 'ATAC':
            latent_layer_R_out, latent_layer_A_out, latent_layer_mu, latent_layer_d = self.forward_with_ATAC(x, 'test')
            return latent_layer_R_out, latent_layer_A_out, latent_layer_mu, latent_layer_d

        
class Single_Translator(nn.Module):
    def __init__(
        self,
        translator_input_dim: int,  
        translator_embed_dim: int, 
        translator_embed_act_list: list, 
        ):
        """
        Single translator block used only for pretraining.
        
        Parameters
        ----------
        translator_input_dim
            dimension of input from encoder for translator.
            
        translator_embed_dim
            dimension of embedding space for translator.
            
        translator_embed_act_list
            activation list for translator, involving [mean_activation, log_var_activation, decoder_activation].
            
        """
        
        super(Single_Translator, self).__init__()
        
        mean_activation, log_var_activation, decoder_activation = translator_embed_act_list
        
        self.encoder_l_mu = nn.Linear(translator_input_dim, translator_embed_dim)
        nn.init.xavier_uniform_(self.encoder_l_mu.weight)
        self.encoder_bn_mu = nn.BatchNorm1d(translator_embed_dim)
        self.encoder_act_mu = mean_activation
        
        self.encoder_l_d = nn.Linear(translator_input_dim, translator_embed_dim)
        nn.init.xavier_uniform_(self.encoder_l_d.weight)
        self.encoder_bn_d = nn.BatchNorm1d(translator_embed_dim)
        self.encoder_act_d = log_var_activation

        self.decoder_l = nn.Linear(translator_embed_dim, translator_input_dim)
        nn.init.xavier_uniform_(self.decoder_l.weight)
        self.decoder_bn = nn.BatchNorm1d(translator_input_dim)
        self.decoder_act = decoder_activation                
    
    def reparameterize(self, mu, sigma):
        sigma = torch.exp(sigma/2)
        eps = torch.randn_like(sigma)
        return mu + eps * sigma
    
    def forward(self, x, forward_type):
            
        latent_layer_mu = self.encoder_act_mu(self.encoder_bn_mu(self.encoder_l_mu(x)))
        latent_layer_d = self.encoder_act_d(self.encoder_bn_d(self.encoder_l_d(x)))
        
        if forward_type == 'test':
            latent_layer = latent_layer_mu
        elif forward_type == 'train':
            latent_layer = self.reparameterize(latent_layer_mu, latent_layer_d)
        
        latent_layer_out = self.decoder_act(self.decoder_bn(self.decoder_l(latent_layer)))
        return latent_layer_out, latent_layer_mu, latent_layer_d



import random
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.utils.data import Dataset
from matplotlib.backends.backend_pdf import PdfPages
import matplotlib.pyplot as plt
from scipy import sparse
import anndata as ad
import pandas as pd
import gc


class RNA_ATAC_dataset(Dataset):
    def __init__(self, R_data, A_data, id_list_r, id_list_a):
        """
        Set random seed.

        Parameters
        ----------
        R_data
            complete RNA data for model traning and testing.
            
        A_data
            complete ATAC data for model traning and testing.

        id_list_r
            ids of cells in RNA data used for model training.
            
        id_list_a
            ids of cells in ATAC data used for model training.    
        
        """
        
        self.RNA_dataset = R_data        
        self.ATAC_dataset = A_data
        self.id_list_r = id_list_r
        self.id_list_a = id_list_a
        self.r_count = len(self.id_list_r)
        self.a_count = len(self.id_list_a)

    def __len__(self):
        return self.r_count

    def __getitem__(self, idx):
        RNA_x = self.RNA_dataset[self.id_list_r[idx], :]
        ATAC_x = self.ATAC_dataset[self.id_list_a[idx], :]
        RNA_x = torch.from_numpy(RNA_x)
        ATAC_x = torch.from_numpy(ATAC_x)
        sample = torch.cat([RNA_x, ATAC_x])
        return sample


class Single_omics_dataset(Dataset):
    def __init__(self, data, id_list):
        """
        Set random seed.

        Parameters
        ----------
        data
            complete data for model traning and testing.

        id_list
            id of cells used for model training.
        
        """

        self.dataset = data        
        self.id_list = id_list

    def __len__(self):
        return len(self.id_list)

    def __getitem__(self, idx):
        x = self.dataset[self.id_list[idx], :]
        x = torch.from_numpy(x)
        return x   


def setup_seed(seed):
    """
    Set random seed.
    
    Parameters
    ----------
    seed
        Number to be set as random seed for reproducibility.
        
    """
    np.random.seed(seed)
    random.seed(seed)
    torch.cuda.manual_seed_all(seed)
    torch.manual_seed(seed)


class EarlyStopping:
    """Cite from https://github.com/Bjarten/early-stopping-pytorch"""
    """Early stops the training if validation loss doesn't improve after a given patience."""
    def __init__(self, patience=7, verbose=False, delta=0):
        """
        Args:
            patience (int): How long to wait after last time validation loss improved.
                            Default: 7
            verbose (bool): If True, prints a message for each validation loss improvement. 
                            Default: False
            delta (float): Minimum change in the monitored quantity to qualify as an improvement.
                            Default: 0
        """
        self.patience = patience
        self.verbose = verbose
        self.counter = 0
        self.best_score = None
        self.early_stop = False
        self.val_loss_min = np.Inf
        self.delta = delta

    def __call__(self, val_loss, model, path):

        score = -val_loss

        if self.best_score is None:
            self.best_score = score
            self.save_checkpoint(val_loss, model, path)
        elif score < self.best_score + self.delta:
            self.counter += 1
            # print(f'EarlyStopping counter: {self.counter} out of {self.patience}')
            if self.counter >= self.patience:
                self.early_stop = True
        else:
            self.best_score = score
            self.save_checkpoint(val_loss, model, path)
            self.counter = 0

    def save_checkpoint(self, val_loss, model, path):
        '''Saves model when validation loss decrease.'''
        if self.verbose:
            print(f'Validation loss decreased ({self.val_loss_min:.6f} --> {val_loss:.6f}).  Saving model ...')
        model.save_model_dict(path) # save best model here
        self.val_loss_min = val_loss

        
def tensor2adata(x, val=1e-4):
    """
    Transform the tensor list to an Anndata.
    
    Parameters
    ----------
    x
        tensor list to concatenate.

    val
        value for sparse for matrix, default 1e-4.
        
    Returns
    ----------
    x
        concatenated anndata.
        
    """
    x = torch.cat(x)
    x = torch.masked_fill(x, x < val, 0)
    x = sparse.csr_matrix(x)
    x = ad.AnnData(x)
    return x


def tensor2adata_adt(x):
    """
    Transform the tensor list to an Anndata.
    
    Parameters
    ----------
    x
        tensor list to concatenate.

    Returns
    ----------
    x
        concatenated anndata.
        
    """
    x = torch.cat(x)
    x = sparse.csr_matrix(x)
    x = ad.AnnData(x)
    return x


def record_loss_log(
    pretrain_r_loss,
    pretrain_r_kl,
    pretrain_r_loss_val,
    pretrain_r_kl_val,
    pretrain_a_loss,
    pretrain_a_kl,
    pretrain_a_loss_val,
    pretrain_a_kl_val,
    train_loss,
    train_kl,
    train_discriminator,
    train_loss_val,
    train_kl_val,
    train_discriminator_val,
    path
):
    """
    Record run loss of training.
    
    Parameters
    ----------
    pretrain_r_loss
        train reconstruct loss of pretrain for RNA.
        
    pretrain_r_kl
        train kl divergence of pretrain for RNA.
        
    pretrain_r_loss_val
        validation reconstruct loss of pretrain for RNA.
        
    pretrain_r_kl_val
        validation kl divergence of pretrain for RNA.
        
    pretrain_a_loss
        train reconstruct loss of pretrain for ATAC.
        
    pretrain_a_kl
        train kl divergence of pretrain for ATAC.
        
    pretrain_a_loss_val
        validation reconstruct loss of pretrain for ATAC.
        
    pretrain_a_kl_val
        validation kl divergence of pretrain for ATAC.
        
    train_loss
        train reconstruct loss of train.
        
    train_kl
        train kl divergence of train.
        
    train_discriminator
        validation reconstruct loss of train.
        
    train_loss_val
        train reconstruct loss of train.
        
    train_kl_val
        train kl divergence of train.
        
    train_discriminator_val
        validation reconstruct train.
   
    path
        path for saving runlog.
    
    """
    fig_pretrain_r = plt.figure()
    ax_pretrain_r_re = fig_pretrain_r.add_subplot(1, 2, 1)
    ax_pretrain_r_kl = fig_pretrain_r.add_subplot(1, 2, 2)
    
    ax_pretrain_r_re.plot(pretrain_r_loss)
    ax_pretrain_r_re.plot(pretrain_r_loss_val)
    ax_pretrain_r_re.set_title('Reconstruct Loss for RNA Pretrain')
    
    ax_pretrain_r_kl.plot(pretrain_r_kl)
    ax_pretrain_r_kl.plot(pretrain_r_kl_val)
    ax_pretrain_r_kl.set_title('KL Divergence for RNA Pretrain')
    fig_pretrain_r.tight_layout()
    
    fig_pretrain_a = plt.figure()   
    ax_pretrain_a_re = fig_pretrain_a.add_subplot(1, 2, 1)
    ax_pretrain_a_kl = fig_pretrain_a.add_subplot(1, 2, 2)
    
    ax_pretrain_a_re.plot(pretrain_a_loss)
    ax_pretrain_a_re.plot(pretrain_a_loss_val)
    ax_pretrain_a_re.set_title('Reconstruct Loss for ATAC Pretrain')
    
    ax_pretrain_a_kl.plot(pretrain_a_kl)
    ax_pretrain_a_kl.plot(pretrain_a_kl_val)
    ax_pretrain_a_kl.set_title('KL Divergence for ATAC Pretrain')
    fig_pretrain_a.tight_layout()
    
    fig_train = plt.figure()
    ax_train_re = fig_train.add_subplot(1, 3, 1)
    ax_train_kl = fig_train.add_subplot(1, 3, 2)
    ax_train_dis = fig_train.add_subplot(1, 3, 3)
    
    ax_train_re.plot(train_loss)
    ax_train_re.plot(train_loss_val)
    ax_train_re.set_title('Reconstruct Loss for Train')
    
    ax_train_kl.plot(train_kl)
    ax_train_kl.plot(train_kl_val)
    ax_train_kl.set_title('KL Divergence for Train')
    
    ax_train_dis.plot(train_discriminator)
    ax_train_dis.plot(train_discriminator_val)
    ax_train_dis.set_title('Discriminator Loss for Train')
    fig_train.tight_layout()
    
    fig_list = [fig_pretrain_r, fig_pretrain_a, fig_train]
    with PdfPages(path + '/run_log.pdf') as pdf:
        for i in range(len(fig_list)):
            pdf.savefig(figure=fig_list[i], dpi=200)
            plt.close()
            

def idx2adata_multiVI(
    adata,
    train_id,
    validation_id, 
    test_id
):
    """
    Split datasets into train, validation and test part using cell id.
    
    Parameters
    ----------
    RNA_data
        full RNA data for spliting.
        
    ATAC_data
        full ATAC data for spliting.
        
    train_id
        cell index list used for model training.
   
    validation_id
        cell index list used for model validation.
        
    test_id
        cell index list used for model testing.
        
    """
    #train_id.extend(validation_id)
    train_adata = ad.AnnData(adata.X[train_id, :])
    test_adata = ad.AnnData(adata.X[test_id, :])

    train_adata.var = adata.var
    test_adata.var = adata.var

    train_adata.obs = adata.obs.iloc[train_id, :].copy()
    test_adata.obs = adata.obs.iloc[test_id, :].copy()

    train_adata.obs.reset_index(drop=True, inplace=True)
    test_adata.obs.reset_index(drop=True, inplace=True)

    return train_adata, test_adata


from . import butterfly
from . import data_processing
from . import split_datasets
from . import draw_cluster
from . import train_model
from . import train_model_cite
from . import train_model_perturb
from .version import __version__


import logging

def create_logger(name='', ch=True, fh=False, levelname=logging.INFO, overwrite=False):
    """
    Generate logger 
    
    Parameters
    ----------
    name
        name of logger, default "".
        
    ch
        if True, add console handler output logging to console, default True.
        
    fh
        if True, add file handler output logging to file, default False.
        
    levelname
        level of logger, default logging.INFO.
        
    overwrite
        if True, overwrite the exist handler in current logger, default False.
        
    Return
    ----------
    logger
        logger generated with desired handler, logging level and name.
        
    """

    logger = logging.getLogger(name)
    logger.setLevel(levelname)
    
    if overwrite:
        for h in logger.handlers:
            logger.removeHandler(h)
    
    formatter = logging.Formatter('[%(levelname)s] %(name)s: %(message)s')

    if ch:
        ch = logging.StreamHandler()
        ch.setLevel(logging.INFO)
        ch.setFormatter(formatter)
        logger.addHandler(ch)
    if fh:
        fh = logging.FileHandler(fh, mode='w')
        fh.setLevel(logging.DEBUG)
        fh.setFormatter(formatter)
        logger.addHandler(fh)
    return logger


import getopt
import sys
import gc
import os
from scButterfly.data_processing import RNA_data_preprocessing, ATAC_data_preprocessing
import scanpy as sc
import anndata as ad
from scButterfly.logger import *
from scButterfly.model_utlis import *
import warnings
import torch
import torch.nn as nn
from scButterfly.train_model import Model
import torch
import torch.nn as nn
warnings.filterwarnings("ignore")

class Butterfly():
    def __init__(
        self,
     ):
        """
        Butterfly model.

        """
        setup_seed(19193)
        self.my_logger = create_logger(name='Butterfly', ch=True, fh=False, levelname=logging.INFO, overwrite=False)
    
    def load_data(
        self,
        RNA_data,
        ATAC_data,
        train_id,
        test_id,
        validation_id = None,
     ):
        """
        Load data to Butterfly model.
        
        Parameters
        ----------
        train_id: list
            list of cell ids for training.
            
        test_id: list
            list of cell ids for testing.
            
        validation_id: list
            list of cell ids for validation, if None, Butterfly will use default 20% train cells for validation.

        """
        self.RNA_data = RNA_data.copy()
        self.ATAC_data = ATAC_data.copy()
        self.test_id_r = test_id.copy()
        self.test_id_a = test_id.copy()
        if validation_id is None:
            self.train_id = train_id
            random.shuffle(self.train_id)
            train_count = int(len(self.train_id)*0.8)
            self.train_id_r = self.train_id[0:train_count].copy()
            self.train_id_a = self.train_id[0:train_count].copy()
            self.validation_id_r = self.train_id[train_count:].copy()
            self.validation_id_a = self.train_id[train_count:].copy()
            del self.train_id
        else:
            self.train_id_r = train_id.copy()
            self.train_id_a = train_id.copy()
            self.validation_id_r = validation_id.copy()
            self.validation_id_a = validation_id.copy()
        self.my_logger.info('successfully load in data with\n\nRNA data:\n'+str(RNA_data)+'\n\nATAC data:\n'+str(ATAC_data))
        self.is_processed = False
        
    def data_preprocessing(
        self,
        normalize_total=True,
        log1p=True,
        use_hvg=True,
        n_top_genes=3000,
        binary_data=True,
        filter_features=True,
        fpeaks=0.005,
        tfidf=True,
        normalize=True,
        save_data=False,
        file_path=None,
        logging_path=None
    ):
        """
        Preprocessing for RNA data and ATAC data in Butterfly.

        Parameters
        ----------
        normalize_total: bool
            choose use normalization or not, default True.

        log1p: bool
            choose use log transformation or not, default True.

        use_hvg: bool
            choose use highly variable genes or not, default True.

        n_top_genes: int
            the count of highly variable genes, if not use highly variable, set use_hvg = False and n_top_genes = None, default 3000.
            
        binary_data: bool
            choose binarized ATAC data or not, default True.

        filter_features: bool
            choose use peaks filtering or not, default True.

        fpeaks: float
            the fpeaks for peaks filtering, is don't filter peaks set it None, default 0.005.

        tfidf: bool
            choose using TF-IDF transform or not, default True.

        normalize: bool
            choose set data to [0, 1] or not, default True.

        save_data: bool
            choose save the processed data or not, default False.

        file_path: str
            the path for saving processed data, only used if save_data is True, default None.

        logging_path: str
            the path for output process logging, if not save, set it None, default None.
            
         """
        if self.is_processed:
            self.my_logger.warning('finding data have already been processed!')
        else:
            self.RNA_data_p = RNA_data_preprocessing(
                self.RNA_data,
                normalize_total=normalize_total,
                log1p=log1p,
                use_hvg=use_hvg,
                n_top_genes=n_top_genes,
                save_data=save_data,
                file_path=file_path,
                logging_path=file_path
                )
            self.ATAC_data_p = ATAC_data_preprocessing(
                self.ATAC_data,
                binary_data=binary_data,
                filter_features=filter_features,
                fpeaks=fpeaks,
                tfidf=tfidf,
                normalize=normalize,
                save_data=save_data,
                file_path=file_path,
                logging_path=file_path
            )[0]
            self.is_processed = True
            
    def amplification(
        self,
        amp_type=None,
        MultiVI_path=None
    ):
        """
        Data amplification for Butterfly model.
        
        Parameters
        ----------
        amp_type: str
            Butterfly support two types of amp_type, "cell_type_amplification" and "MultiVI_amplification". "cell_type_amplification" need "cell_type" in RNA_data.obs and ATAC_data.obs， default None.
            
        MultiVI_path: str
            path for pretrained MultiVI model, if None, Butterfly will train a MultiVI model first, default None.

        """
        if amp_type == 'cell_type_amplification' and 'cell_type' in self.RNA_data.obs.keys():
            self.my_logger.info('using data amplification with cell type labels.')
            copy_count = 3
            random.seed(19193)
            self.ATAC_data.obs.index = [str(i) for i in range(len(self.ATAC_data.obs.index))]
            cell_type = self.ATAC_data.obs.cell_type.iloc[self.train_id_a]
            for i in range(len(cell_type.cat.categories)):
                cell_type_name = cell_type.cat.categories[i]
                idx_temp = list(cell_type[cell_type == cell_type_name].index.astype(int))
                for j in range(copy_count - 1):
                    random.shuffle(idx_temp)
                    self.train_id_r.extend(idx_temp)
                    random.shuffle(idx_temp)
                    self.train_id_a.extend(idx_temp)
        if amp_type == 'cell_type_amplification' and not 'cell_type' in self.RNA_data.obs.keys():
            self.my_logger.warning('not find "cell_type" in data.obs, trying to use MultiVI amplification ...')
            amp_type = 'MultiVI_amplification'
        if amp_type == 'MultiVI_amplification':
            self.my_logger.info('using data amplification with MultiVI cluster labels.')
            from scvi_colab import install
            import pandas as pd
            install()
            import scvi
            import sys
            import scipy.sparse as sp

            adata = ad.AnnData(sp.hstack((self.RNA_data.X, self.ATAC_data.X)))
            adata.X = adata.X.tocsr()
            adata.obs = self.RNA_data.obs

            m = len(self.RNA_data.var.index)
            n = len(self.ATAC_data.var.index)
            adata.var.index = pd.Series([self.RNA_data.var.index[i] if i<m else self.ATAC_data.var.index[i-m] for i in range(m+n)], dtype='object')
            adata.var['modality'] = pd.Series(['Gene Expression' if i<m else 'Peaks' for i in range(m+n)], dtype='object').values

            adata.var_names_make_unique()

            adata_mvi = scvi.data.organize_multiome_anndatas(adata)

            adata_mvi = adata_mvi[:, adata_mvi.var["modality"].argsort()].copy()

            sc.pp.filter_genes(adata_mvi, min_cells=int(adata_mvi.shape[0] * 0.01))

            train_id, validation_id, test_id = self.train_id_r, self.validation_id_r, self.test_id_r

            train_adata = idx2adata_multiVI(adata_mvi, train_id, validation_id, test_id)[0]
            
            if MultiVI_path is None:
                self.my_logger.info('no trained model find, train MultiVI model first ...')
                
                scvi.model.MULTIVI.setup_anndata(train_adata, batch_key='modality')

                mvi = scvi.model.MULTIVI(
                    train_adata,
                    n_genes=(adata_mvi.var['modality']=='Gene Expression').sum(),
                    n_regions=(adata_mvi.var['modality']=='Peaks').sum(),
                )

                mvi.train()
            else:
                self.my_logger.info('load trained MultiVI model from '+MultiVI_path)
                mvi = scvi.model.MULTIVI.load(MultiVI_path, adata=train_adata)

            train_adata.obsm["MultiVI_latent"] = mvi.get_latent_representation()
            leiden_adata = ad.AnnData(train_adata.obsm["MultiVI_latent"])
            sc.pp.neighbors(leiden_adata)
            sc.tl.leiden(leiden_adata, resolution=3)

            del train_adata, adata_mvi
            gc.collect()
            
            
            copy_count = 3
            random.seed(19193)
            cell_type = leiden_adata.obs.leiden
            for i in range(len(cell_type.cat.categories)):
                cell_type_name = cell_type.cat.categories[i]
                idx_temp = list(cell_type[cell_type == cell_type_name].index.astype(int))
                for j in range(copy_count - 1):
                    random.shuffle(idx_temp)
                    for each in idx_temp:
                        self.train_id_r.append(train_id[each])
                    random.shuffle(idx_temp)
                    for each in idx_temp:
                        self.train_id_a.append(train_id[each])

    def construct_model(
        self,
        chrom_list,
        logging_path = None,
        R_encoder_nlayer = 2, 
        A_encoder_nlayer = 2,
        R_decoder_nlayer = 2, 
        A_decoder_nlayer = 2,
        R_encoder_dim_list = [256, 128],
        A_encoder_dim_list = [32, 128],
        R_decoder_dim_list = [128, 256],
        A_decoder_dim_list = [128, 32],
        R_encoder_act_list = [nn.LeakyReLU(), nn.LeakyReLU()],
        A_encoder_act_list = [nn.LeakyReLU(), nn.LeakyReLU()],
        R_decoder_act_list = [nn.LeakyReLU(), nn.LeakyReLU()],
        A_decoder_act_list = [nn.LeakyReLU(), nn.Sigmoid()],
        translator_embed_dim = 128, 
        translator_input_dim_r = 128,
        translator_input_dim_a = 128,
        translator_embed_act_list = [nn.LeakyReLU(), nn.LeakyReLU(), nn.LeakyReLU()],
        discriminator_nlayer = 1,
        discriminator_dim_list_R = [128],
        discriminator_dim_list_A = [128],
        discriminator_act_list = [nn.Sigmoid()],
        dropout_rate = 0.1,
        R_noise_rate = 0.5,
        A_noise_rate = 0.3,
    ):
        """
        Main model.
        
        Parameters
        ----------
        chrom_list: list
            list of peaks count for each chromosomes.
            
        logging_path: str
            the path for output process logging, if not save, set it None, default None.

        R_encoder_nlayer: int
            layer counts of RNA encoder, default 2.
            
        A_encoder_nlayer: int
            layer counts of ATAC encoder, default 2.
            
        R_decoder_nlayer: int
            layer counts of RNA decoder, default 2.
            
        A_decoder_nlayer: int
            layer counts of ATAC decoder, default 2.
            
        R_encoder_dim_list: list
            dimension list of RNA encoder, length equal to R_encoder_nlayer, default [256, 128].
            
        A_encoder_dim_list: list
            dimension list of ATAC encoder, length equal to A_encoder_nlayer, default [32, 128].
            
        R_decoder_dim_list: list
            dimension list of RNA decoder, length equal to R_decoder_nlayer, default [128, 256].
            
        A_decoder_dim_list: list
            dimension list of ATAC decoder, length equal to A_decoder_nlayer, default [128, 32].
            
        R_encoder_act_list: list
            activation list of RNA encoder, length equal to R_encoder_nlayer, default [nn.LeakyReLU(), nn.LeakyReLU()].
            
        A_encoder_act_list: list
            activation list of ATAC encoder, length equal to A_encoder_nlayer, default [nn.LeakyReLU(), nn.LeakyReLU()].
            
        R_decoder_act_list: list
            activation list of RNA decoder, length equal to R_decoder_nlayer, default [nn.LeakyReLU(), nn.LeakyReLU()].
            
        A_decoder_act_list: list
            activation list of ATAC decoder, length equal to A_decoder_nlayer, default [nn.LeakyReLU(), nn.Sigmoid()].
            
        translator_embed_dim: int
            dimension of embedding space for translator, default 128.
            
        translator_input_dim_r: int
            dimension of input from RNA encoder for translator, default 128.
            
        translator_input_dim_a: int
            dimension of input from ATAC encoder for translator, default 128.
            
        translator_embed_act_list: list
            activation list for translator, involving [mean_activation, log_var_activation, decoder_activation], default [nn.LeakyReLU(), nn.LeakyReLU(), nn.LeakyReLU()].
            
        discriminator_nlayer: int
            layer counts of discriminator, default 1.
            
        discriminator_dim_list_R: list
            dimension list of discriminator, length equal to discriminator_nlayer, the first equal to translator_input_dim_R, default [128].
            
        discriminator_dim_list_A: list
            dimension list of discriminator, length equal to discriminator_nlayer, the first equal to translator_input_dim_A, default [128].
            
        discriminator_act_list: list
            activation list of discriminator, length equal to  discriminator_nlayer, default [nn.Sigmoid()].
            
        dropout_rate: float
            rate of dropout for network, default 0.1.
       
        R_noise_rate: float
            rate of set part of RNA input data to 0, default 0.5.
            
        A_noise_rate: float
            rate of set part of ATAC input data to 0, default 0.3.

        """
        R_encoder_dim_list.insert(0, self.RNA_data_p.X.shape[1]) 
        A_encoder_dim_list.insert(0, self.ATAC_data_p.X.shape[1])
        R_decoder_dim_list.append(self.RNA_data_p.X.shape[1]) 
        A_decoder_dim_list.append(self.ATAC_data_p.X.shape[1])
        A_encoder_dim_list[1] *= len(chrom_list)
        A_decoder_dim_list[1] *= len(chrom_list)
        self.model = Model(
            R_encoder_nlayer = R_encoder_nlayer, 
            A_encoder_nlayer = A_encoder_nlayer,
            R_decoder_nlayer = R_decoder_nlayer, 
            A_decoder_nlayer = A_decoder_nlayer,
            R_encoder_dim_list = R_encoder_dim_list,
            A_encoder_dim_list = A_encoder_dim_list,
            R_decoder_dim_list = R_decoder_dim_list,
            A_decoder_dim_list = A_decoder_dim_list,
            R_encoder_act_list = R_encoder_act_list,
            A_encoder_act_list = A_encoder_act_list,
            R_decoder_act_list = R_decoder_act_list,
            A_decoder_act_list = A_decoder_act_list,
            translator_embed_dim = translator_embed_dim, 
            translator_input_dim_r = translator_input_dim_r,
            translator_input_dim_a = translator_input_dim_a,
            translator_embed_act_list = translator_embed_act_list,
            discriminator_nlayer = discriminator_nlayer,
            discriminator_dim_list_R = discriminator_dim_list_R,
            discriminator_dim_list_A = discriminator_dim_list_A,
            discriminator_act_list = discriminator_act_list,
            dropout_rate = dropout_rate,
            R_noise_rate = R_noise_rate,
            A_noise_rate = A_noise_rate,
            chrom_list = chrom_list,
            logging_path = logging_path,
            RNA_data = self.RNA_data_p,
            ATAC_data = self.ATAC_data_p
        )
        self.my_logger.info('successfully construct butterfly model.')

    def train_model(
        self,
        R_encoder_lr = 0.001,
        A_encoder_lr = 0.001,
        R_decoder_lr = 0.001,
        A_decoder_lr = 0.001,
        R_translator_lr = 0.001,
        A_translator_lr = 0.001,
        translator_lr = 0.001,
        discriminator_lr = 0.005,
        R2R_pretrain_epoch = 100,
        A2A_pretrain_epoch = 100,
        lock_encoder_and_decoder = False,
        translator_epoch = 200,
        patience = 50,
        batch_size = 64,
        r_loss = nn.MSELoss(size_average=True),
        a_loss = nn.BCELoss(size_average=True),
        d_loss = nn.BCELoss(size_average=True),
        loss_weight = [1, 2, 1],
        output_path = None,
        seed = 19193,
        kl_mean = True,
        R_pretrain_kl_warmup = 50,
        A_pretrain_kl_warmup = 50,
        translation_kl_warmup = 50,
        load_model = None,
        logging_path = None
    ):
        """
        Training for model.
        
        Parameters
        ----------
        R_encoder_lr: float
            learning rate of RNA encoder, default 0.001.
            
        A_encoder_lr: float
            learning rate of ATAC encoder, default 0.001.
            
        R_decoder_lr: float
            learning rate of RNA decoder, default 0.001.
            
        A_decoder_lr: float
            learning rate of ATAC decoder, default 0.001.
       
        R_translator_lr: float
            learning rate of RNA pretrain translator, default 0.001.
            
        A_translator_lr: float
            learning rate of ATAC pretrain translator, default 0.001.
            
        translator_lr: float
            learning rate of translator, default 0.001.
            
        discriminator_lr: float
            learning rate of discriminator, default 0.005.
            
        R2R_pretrain_epoch: int
            max epoch for pretrain RNA autoencoder, default 100.
            
        A2A_pretrain_epoch: int
            max epoch for pretrain ATAC autoencoder, default 100.
            
        lock_encoder_and_decoder: bool
            lock the pretrained encoder and decoder or not, default False.
            
        translator_epoch: int
            max epoch for train translator, default 200.
            
        patience: int
            patience for loss on validation, default 50.
            
        batch_size: int
            batch size for training and validation, default 64.
            
        r_loss
            loss function for RNA reconstruction, default nn.MSELoss(size_average=True).
            
        a_loss
            loss function for ATAC reconstruction, default nn.BCELoss(size_average=True).
            
        d_loss
            loss function for discriminator, default nn.BCELoss(size_average=True).
            
        loss_weight: list
            list of loss weight for [r_loss, a_loss, d_loss], default [1, 2, 1].

        output_path: str
            file path for model output, default None.
            
        seed: int
            set up the random seed, default 19193.
            
        kl_mean: bool
            size average for kl divergence or not, default True.
            
        R_pretrain_kl_warmup: int
            epoch of linear weight warm up for kl divergence in RNA pretrain, default 50.
        
        A_pretrain_kl_warmup: int
            epoch of linear weight warm up for kl divergence in ATAC pretrain, default 50.
        
        translation_kl_warmup: int
            epoch of linear weight warm up for kl divergence in translator pretrain, default 50.
            
        load_model: str
            the path for loading model if needed, else set it None, default None.
            
        logging_path: str
            the path for output process logging, if not save, set it None, default None.
            
        """
        self.my_logger.info('training butterfly model ...')
        R_kl_div = 1 / self.RNA_data_p.X.shape[1] * 20
        A_kl_div = 1 / self.ATAC_data_p.X.shape[1] * 20
        kl_div = R_kl_div + A_kl_div
        loss_weight.extend([R_kl_div, A_kl_div, kl_div])
        self.model.train(
            R_encoder_lr = R_encoder_lr,
            A_encoder_lr = A_encoder_lr,
            R_decoder_lr = R_decoder_lr,
            A_decoder_lr = A_decoder_lr,
            R_translator_lr = R_translator_lr,
            A_translator_lr = A_translator_lr,
            translator_lr = translator_lr,
            discriminator_lr = discriminator_lr,
            R2R_pretrain_epoch = R2R_pretrain_epoch,
            A2A_pretrain_epoch = A2A_pretrain_epoch,
            lock_encoder_and_decoder = lock_encoder_and_decoder,
            translator_epoch = translator_epoch,
            patience = patience,
            batch_size = batch_size,
            r_loss = r_loss,
            a_loss = a_loss,
            d_loss = d_loss,
            loss_weight = loss_weight,
            train_id_r = self.train_id_r,
            train_id_a = self.train_id_a,
            validation_id_r = self.validation_id_r, 
            validation_id_a = self.validation_id_a, 
            output_path = output_path,
            seed = seed,
            kl_mean = kl_mean,
            R_pretrain_kl_warmup = R_pretrain_kl_warmup,
            A_pretrain_kl_warmup = A_pretrain_kl_warmup,
            translation_kl_warmup = translation_kl_warmup,
            load_model = load_model,
            logging_path = logging_path
        )

    def test_model(
        self, 
        model_path = None,
        load_model = False,
        output_path = None,
        test_relation = False,
        test_cluster = False,
        test_figure = False,
        output_data = False
    ):
        """
        Test for model.
        
        Parameters
        ----------
        model_path: str
            path for load trained model, default None.
            
        load_model: bool
            load the pretrained model or not, default False.
            
        output_path: str
            file path for model output, default None.
            
        test_relation: bool
            test correlation or not, default False.
            
        test_cluster: bool
            test clustrer index or not, default False.
            
        test_figure: bool
            test tSNE or not, default False.
            
        output_data: bool
            output the predicted test data to file or not, default False.
            
        """
        self.my_logger.info('testing butterfly model ...')
        A2R_predict, R2A_predict = self.model.test(
            test_id_r = self.test_id_r,
            test_id_a = self.test_id_a, 
            model_path = model_path,
            load_model = load_model,
            output_path = output_path,
            test_evaluate = test_relation,
            test_cluster = test_cluster,
            test_figure = test_figure,
            output_data = output_data,
            return_predict = True
        )
        return A2R_predict, R2A_predict


import scanpy as sc
import anndata as ad
import pandas as pd
import torch
import numpy as np
import random
import episcanpy.api as epi
import scipy
from scipy.sparse import csr_matrix
from scButterfly.logger import *


def TFIDF(count_mat): 
    """
    TF-IDF transformation for matrix.

    Parameters
    ----------
    count_mat
        numpy matrix with cells as rows and peak as columns, cell * peak.

    Returns
    ----------
    tfidf_mat
        matrix after TF-IDF transformation.

    divide_title
        matrix divided in TF-IDF transformation process, would be used in "inverse_TFIDF".

    multiply_title
        matrix multiplied in TF-IDF transformation process, would be used in "inverse_TFIDF".

    """

    count_mat = count_mat.T
    divide_title = np.tile(np.sum(count_mat,axis=0), (count_mat.shape[0],1))
    nfreqs = 1.0 * count_mat / divide_title
    multiply_title = np.tile(np.log(1 + 1.0 * count_mat.shape[1] / np.sum(count_mat,axis=1)).reshape(-1,1), (1,count_mat.shape[1]))
    tfidf_mat = scipy.sparse.csr_matrix(np.multiply(nfreqs, multiply_title)).T
    return tfidf_mat, divide_title, multiply_title


def inverse_TFIDF(TDIDFed_mat, divide_title, multiply_title, max_temp): 
    """
    Inversed TF-IDF transformation for matrix.

    Parameters
    ----------
    TDIDFed_mat: csr_matrix
        matrix after TFIDF transformation with peaks as rows and cells as columns, peak * cell.
        
    divide_title: numpy matrix
        matrix divided in TF-IDF transformation process, could get from "ATAC_data_preprocessing".
        
    multiply_title: numpy matrix
        matrix multiplied in TF-IDF transformation process, could get from "ATAC_data_preprocessing".
        
    max_temp: float
        max scale factor divided in ATAC preprocessing, could get from "ATAC_data_preprocessing".
        
    Returns
    ----------
    count_mat: csr_matrix
        recovered count matrix from matrix after TFIDF transformation.
    """
    
    count_mat = TDIDFed_mat.T
    count_mat = count_mat * max_temp
    nfreqs = np.divide(count_mat, multiply_title)
    count_mat = np.multiply(nfreqs, divide_title).T
    return count_mat


def RNA_data_preprocessing(
    RNA_data,
    normalize_total = True,
    log1p = True,
    use_hvg = True,
    n_top_genes = 3000,
    save_data = False,
    file_path = None,
    logging_path = None
):
    """
    Preprocessing for RNA data, we choose normalization, log transformation and highly variable genes, using scanpy.
    
    Parameters
    ----------
    RNA_data: Anndata
        RNA anndata for processing.
        
    normalize_total: bool
        choose use normalization or not, default True.
        
    log1p: bool
        choose use log transformation or not, default True.
        
    use_hvg: bool
        choose use highly variable genes or not, default True.
        
    n_top_genes: int
        the count of highly variable genes, if not use highly variable, set use_hvg = False and n_top_genes = None, default 3000.
        
    save_data: bool
        choose save the processed data or not, default False.
        
    file_path: str
        the path for saving processed data, only used if save_data is True, default None.
   
    logging_path: str
        the path for output process logging, if not save, set it None, default None.

    Returns
    ---------
    RNA_data_processed: Anndata
        RNA data with normalization, log transformation and highly variable genes selection preprocessed.
    """
    RNA_data_processed = RNA_data.copy()
    
    my_logger = create_logger(name='RNA preprocessing', ch=True, fh=False, levelname=logging.INFO, overwrite=False)
    if not logging_path is None:
        file_handle=open(logging_path + '/Parameters_Record.txt',mode='w')
        file_handle.writelines([
            '------------------------------\n',
            'RNA Preprocessing\n',
            'normalize_total: '+str(normalize_total)+'\n', 
            'log1p: '+str(log1p)+'\n',
            'use_hvg: '+str(use_hvg)+'\n', 
            'n_top_genes: '+str(n_top_genes)+'\n'
        ])
        file_handle.close()
    
    RNA_data_processed.var_names_make_unique()
    
    if not normalize_total or not log1p or not use_hvg:
        my_logger.warning('prefered to process data with default settings to keep the best result.')
    
    if normalize_total:
        my_logger.info('normalize size factor.')
        sc.pp.normalize_total(RNA_data_processed)
        
    if log1p:
        my_logger.info('log transform RNA data.')
        sc.pp.log1p(RNA_data_processed)
    
    if use_hvg:
        my_logger.info('choose top '+str(n_top_genes)+' genes for following training.')
        sc.pp.highly_variable_genes(RNA_data_processed, n_top_genes=n_top_genes)
        RNA_data_processed = RNA_data_processed[:, RNA_data_processed.var['highly_variable']]
    
    if save_data:
        my_logger.warning('writing processed RNA data to target file.')
        if use_hvg:
            RNA_data_processed.write_h5ad(file_path + '/normalize_' + str(normalize_total) + '_log1p_' + str(log1p) + '_hvg_' + str(use_hvg) + '_' + str(n_top_genes) + '_RNA_processed_data.h5ad')
        else:
            RNA_data_processed.write_h5ad(file_path + '/normalize_' + str(normalize_total) + '_log1p_' + str(log1p) + '_hvg_' + str(use_hvg) + '_RNA_processed_data.h5ad')
            
    return RNA_data_processed


def ATAC_data_preprocessing(
    ATAC_data,
    binary_data = True,
    filter_features = True,
    fpeaks = 0.005,
    tfidf = True,
    normalize = True,
    save_data = False,
    file_path = None,
    logging_path = None
):
    """
    Preprocessing for ATAC data, we choose binarize, peaks filtering, TF-IDF transformation and scale transformation, using scanpy.
    
    Parameters
    ----------
    ATAC_data: Anndata
        ATAC anndata for processing.
        
    binary_data: bool
        choose binarized ATAC data or not, default True.
        
    filter_features: bool
        choose use peaks filtering or not, default True.
        
    fpeaks: float
        the fpeaks for peaks filtering, is don't filter peaks set it None, default 0.005.
        
    tfidf: bool
        choose using TF-IDF transform or not, default True.
    
    normalize: bool
        choose set data to [0, 1] or not, default True.
        
    save_data: bool
        choose save the processed data or not, default False.
        
    file_path: str
        the path for saving processed data, only used if save_data is True, default None.
   
    logging_path: str
        the path for output process logging, if not save, set it None, default None.

    Returns
    ---------
    ATAC_data_processed: Anndata 
        ATAC data with binarization, peaks filtering, TF-IDF transformation and scale transformation preprocessed.

    divide_title: numpy matrix
        matrix divided in TF-IDF transformation process, would be used in "inverse_TFIDF".

    multiply_title: numpy matrix
        matrix multiplied in TF-IDF transformation process, would be used in "inverse_TFIDF".
        
    max_temp: float
        max scale factor divided in process, would be used in "inverse_TFIDF".
        
    """
    ATAC_data_processed = ATAC_data.copy()
    
    my_logger = create_logger(name='ATAC preprocessing', ch=True, fh=False, levelname=logging.INFO, overwrite=False)
    if not logging_path is None:
        file_handle=open(logging_path + '/Parameters_Record.txt',mode='a')
        file_handle.writelines([
            '------------------------------\n'
            'ATAC Preprocessing\n'
            'binary_data: '+str(binary_data)+'\n', 
            'filter_features: '+str(filter_features)+'\n',
            'fpeaks: '+str(fpeaks)+'\n', 
            'tfidf: '+str(tfidf)+'\n',
            'normalize: '+str(normalize)+'\n'
        ])
        file_handle.close()
    
    if not binary_data or not filter_features or not tfidf or not normalize:
        my_logger.warning('prefered to process data with default settings to keep the best result.')
    
    if binary_data:
        my_logger.info('binarizing data.')
        epi.pp.binarize(ATAC_data_processed)
        
    if filter_features:
        my_logger.info('filter out peaks appear lower than ' + str(fpeaks * 100) + '% cells.')
        epi.pp.filter_features(ATAC_data_processed, min_cells=np.ceil(fpeaks*ATAC_data.shape[0]))

    if tfidf:
        my_logger.info('TF-IDF transformation.')
        count_mat = ATAC_data_processed.X.copy()
        ATAC_data_processed.X, divide_title, multiply_title = TFIDF(count_mat)
    
    if normalize:
        my_logger.info('normalizing data.')
        max_temp = np.max(ATAC_data_processed.X)
        ATAC_data_processed.X = ATAC_data_processed.X / max_temp
    
    if save_data:
        my_logger.warning('writing processed ATAC data to target file.')
        if filter_features:
            ATAC_data.write_h5ad(file_path + '/binarize_' + str(binary_data) + '_filter_' + str(filter_features) + '_fpeaks_' + str(fpeaks)  + '_tfidf_' + str(tfidf) + '_normalize_' + str(normalize) + '_ATAC_processed_data.h5ad')
        else:
            ATAC_data.write_h5ad(file_path + '/binarize_' + str(binary_data) + '_filter_' + str(filter_features) + '_tfidf_' + str(tfidf) + '_normalize_' + str(normalize) + '_ATAC_processed_data.h5ad')

    return ATAC_data_processed, divide_title, multiply_title, max_temp


from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setup(
    name="scButterfly",
    version="0.0.1",
    description="Single-cell cross-modality translation via multi-use dual-aligned variational autoencoders",
    classifiers=[
        "Programming Language :: Python :: 3.9",
    ],
    keywords="single cell, cross-modality translation, dual-aligned variational autoencoder",
    packages=find_packages(),
    python_requires=">=3.9",
    install_requires=requirements,
)


import os
import random
import time
import numpy as np
import scanpy as sc
import pandas as pd
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.utils.data import Dataset, DataLoader
from tqdm import tqdm
from scButterfly.model_component import *
from scButterfly.model_utlis import *
from scButterfly.calculate_cluster import *
from scButterfly.draw_cluster import *
from scButterfly.relation_evaluation import *
from scButterfly.data_processing import *
from scButterfly.logger import *


class Model():
    def __init__(self,
        RNA_data,
        ATAC_data,        
        chrom_list: list,
        logging_path: str,
        R_encoder_dim_list: list,
        A_encoder_dim_list: list,
        R_decoder_dim_list: list,
        A_decoder_dim_list: list,
        R_encoder_nlayer: int = 2, 
        A_encoder_nlayer: int = 2,
        R_decoder_nlayer: int = 2, 
        A_decoder_nlayer: int = 2,
        R_encoder_act_list: list = [nn.LeakyReLU(), nn.LeakyReLU()],
        A_encoder_act_list: list = [nn.LeakyReLU(), nn.LeakyReLU()],
        R_decoder_act_list: list = [nn.LeakyReLU(), nn.LeakyReLU()],
        A_decoder_act_list: list = [nn.LeakyReLU(), nn.Sigmoid()],
        translator_embed_dim: int = 128, 
        translator_input_dim_r: int = 128,
        translator_input_dim_a: int = 128,
        translator_embed_act_list: list = [nn.LeakyReLU(), nn.LeakyReLU(), nn.LeakyReLU()],
        discriminator_nlayer: int = 1,
        discriminator_dim_list_R: list = [128],
        discriminator_dim_list_A: list = [128],
        discriminator_act_list: list = [nn.Sigmoid()],
        dropout_rate: float = 0.1,
        R_noise_rate: float = 0.5,
        A_noise_rate: float = 0,
    ):
        """
        Main model. Some parameters need information about data, please see in Tutorial.
        
        Parameters
        ----------
        RNA_data: Anndata
            RNA data for model training and testing.
            
        ATAC_data: Anndata
            ADT data for model training and testing.

        chrom_list: list
            list of peaks count for each chromosomes.
            
        logging_path: str
            the path for output process logging, if not save, set it None.
            
        R_encoder_dim_list: list
            dimension list of RNA encoder, length equal to R_encoder_nlayer + 1, the first equal to RNA data dimension, the last equal to embedding dimension.
            
        A_encoder_dim_list: list
            dimension list of ADT encoder, length equal to A_encoder_nlayer + 1, the first equal to RNA data dimension, the last equal to embedding dimension.
            
        R_decoder_dim_list: list
            dimension list of RNA decoder, length equal to R_decoder_nlayer + 1, the last equal to embedding dimension, the first equal to RNA data dimension.
            
        A_decoder_dim_list: list
            dimension list of ADT decoder, length equal to A_decoder_nlayer + 1, the last equal to embedding dimension, the first equal to RNA data dimension.
            
        R_encoder_nlayer: int
            layer counts of RNA encoder, default 2.
            
        A_encoder_nlayer: int
            layer counts of ADT encoder, default 2.
            
        R_decoder_nlayer: int
            layer counts of RNA decoder, default 2.
            
        A_decoder_nlayer: int
            layer counts of ADT decoder, default 2.
            
        R_encoder_act_list: list
            activation list of RNA encoder, length equal to R_encoder_nlayer, default [nn.LeakyReLU(), nn.LeakyReLU()].
            
        A_encoder_act_list: list
            activation list of ADT encoder, length equal to A_encoder_nlayer, default [nn.LeakyReLU(), nn.LeakyReLU()].
            
        R_decoder_act_list: list
            activation list of RNA decoder, length equal to R_decoder_nlayer, default [nn.LeakyReLU(), nn.LeakyReLU()].
            
        A_decoder_act_list: list
            activation list of ADT decoder, length equal to A_decoder_nlayer, default [nn.LeakyReLU(), nn.Sigmoid()].
            
        translator_embed_dim: int
            dimension of embedding space for translator, default 128.
            
        translator_input_dim_r: int
            dimension of input from RNA encoder for translator, default 128.
            
        translator_input_dim_a: int
            dimension of input from ADT encoder for translator, default 128.
            
        translator_embed_act_list: list
            activation list for translator, involving [mean_activation, log_var_activation, decoder_activation], default [nn.LeakyReLU(), nn.LeakyReLU(), nn.LeakyReLU()].
            
        discriminator_nlayer: int
            layer counts of discriminator, default 1.
            
        discriminator_dim_list_R: list
            dimension list of discriminator, length equal to discriminator_nlayer, the first equal to translator_input_dim_R, default [128].
            
        discriminator_dim_list_A: list
            dimension list of discriminator, length equal to discriminator_nlayer, the first equal to translator_input_dim_A, default [128].
            
        discriminator_act_list: list
            activation list of discriminator, length equal to  discriminator_nlayer, default [nn.Sigmoid()].
            
        dropout_rate: float
            rate of dropout for network, default 0.1.
       
        R_noise_rate: float
            rate of set part of RNA input data to 0, default 0.5.
            
        A_noise_rate: float
            rate of set part of ADT input data to 0, default 0.

        """
        if not logging_path is None:
            file_handle=open(logging_path + '/Parameters_Record.txt',mode='a')
            file_handle.writelines([
                '------------------------------\n'
                'Model Parameters\n'
                'R_encoder_nlayer: '+str(R_encoder_nlayer)+'\n', 
                'A_encoder_nlayer: '+str(A_encoder_nlayer)+'\n',
                'R_decoder_nlayer: '+str(R_decoder_nlayer)+'\n', 
                'A_decoder_nlayer: '+str(A_decoder_nlayer)+'\n',
                'R_encoder_dim_list: '+str(R_encoder_dim_list)+'\n',
                'A_encoder_dim_list: '+str(A_encoder_dim_list)+'\n',
                'R_decoder_dim_list: '+str(R_decoder_dim_list)+'\n',
                'A_decoder_dim_list: '+str(A_decoder_dim_list)+'\n',
                'R_encoder_act_list: '+str(R_encoder_act_list)+'\n',
                'A_encoder_act_list: '+str(A_encoder_act_list)+'\n',
                'R_decoder_act_list: '+str(R_decoder_act_list)+'\n',
                'A_decoder_act_list: '+str(A_decoder_act_list)+'\n',
                'translator_embed_dim: '+str(translator_embed_dim)+'\n', 
                'translator_input_dim_r: '+str(translator_input_dim_r)+'\n',
                'translator_input_dim_a: '+str(translator_input_dim_a)+'\n',
                'translator_embed_act_list: '+str(translator_embed_act_list)+'\n',
                'discriminator_nlayer: '+str(discriminator_nlayer)+'\n',
                'discriminator_dim_list_R: '+str(discriminator_dim_list_R)+'\n',
                'discriminator_dim_list_A: '+str(discriminator_dim_list_A)+'\n',
                'discriminator_act_list: '+str(discriminator_act_list)+'\n',
                'dropout_rate: '+str(dropout_rate)+'\n',
                'R_noise_rate: '+str(R_noise_rate)+'\n',
                'A_noise_rate: '+str(A_noise_rate)+'\n',
                'chrom_list: '+str(chrom_list)+'\n'
            ])
            file_handle.close()
        
        self.RNA_encoder = NetBlock(
            nlayer = R_encoder_nlayer,
            dim_list = R_encoder_dim_list,
            act_list = R_encoder_act_list,
            dropout_rate = dropout_rate,
            noise_rate = R_noise_rate)

        self.ATAC_encoder = NetBlock(
            nlayer = A_encoder_nlayer,
            dim_list = A_encoder_dim_list,
            act_list = A_encoder_act_list,
            dropout_rate = dropout_rate,
            noise_rate = A_noise_rate)
            
        self.RNA_decoder = NetBlock(
            nlayer = R_decoder_nlayer,
            dim_list = R_decoder_dim_list,
            act_list = R_decoder_act_list,
            dropout_rate = dropout_rate,
            noise_rate = 0)
            
        self.ATAC_decoder = NetBlock(
            nlayer = A_decoder_nlayer,
            dim_list = A_decoder_dim_list,
            act_list = A_decoder_act_list,
            dropout_rate = dropout_rate,
            noise_rate = 0)
        
        self.R_translator = Single_Translator(
            translator_input_dim = translator_input_dim_r, 
            translator_embed_dim = translator_embed_dim, 
            translator_embed_act_list = translator_embed_act_list)
        
        self.A_translator = Single_Translator(
            translator_input_dim = translator_input_dim_a, 
            translator_embed_dim = translator_embed_dim, 
            translator_embed_act_list = translator_embed_act_list)
        
        self.translator = Translator(
            translator_input_dim_r = translator_input_dim_r, 
            translator_input_dim_a = translator_input_dim_a, 
            translator_embed_dim = translator_embed_dim, 
            translator_embed_act_list = translator_embed_act_list)

        discriminator_dim_list_R.append(1)
        discriminator_dim_list_A.append(1)
        self.discriminator_R = NetBlock(
            nlayer = discriminator_nlayer,
            dim_list = discriminator_dim_list_R,
            act_list = discriminator_act_list,
            dropout_rate = 0,
            noise_rate = 0)

        self.discriminator_A = NetBlock(
            nlayer = discriminator_nlayer,
            dim_list = discriminator_dim_list_A,
            act_list = discriminator_act_list,
            dropout_rate = 0,
            noise_rate = 0)
        
        # use GPU for training if cuda is available
        if torch.cuda.is_available():
            self.RNA_encoder = self.RNA_encoder.cuda()
            self.RNA_decoder = self.RNA_decoder.cuda()
            self.ATAC_encoder = self.ATAC_encoder.cuda()
            self.ATAC_decoder = self.ATAC_decoder.cuda()
            self.R_translator = self.R_translator.cuda()
            self.A_translator = self.A_translator.cuda()
            self.translator = self.translator.cuda()
            self.discriminator_R = self.discriminator_R.cuda()
            self.discriminator_A = self.discriminator_A.cuda()
        
        self.is_train_finished = False
        self.RNA_data_obs = RNA_data.obs
        self.ATAC_data_obs = ATAC_data.obs
        self.RNA_data_var = RNA_data.var
        self.ATAC_data_var = RNA_data.var
        self.RNA_data = RNA_data.X.toarray()
        self.ATAC_data = ATAC_data.X.toarray()
        self.num_workers = 0
        
    def set_train(self):
        self.RNA_encoder.train()
        self.RNA_decoder.train()
        self.ATAC_encoder.train()
        self.ATAC_decoder.train()
        self.R_translator.train()
        self.A_translator.train()
        self.translator.train()
        self.discriminator_R.train()
        self.discriminator_A.train()
     
    
    def set_eval(self):
        self.RNA_encoder.eval()
        self.RNA_decoder.eval()
        self.ATAC_encoder.eval()
        self.ATAC_decoder.eval()
        self.R_translator.eval()
        self.A_translator.eval()        
        self.translator.eval()
        self.discriminator_R.eval()
        self.discriminator_A.eval()
    
        
    def forward_R2R(self, RNA_input, r_loss, kl_div_w, forward_type):
        latent_layer, mu, d = self.R_translator(self.RNA_encoder(RNA_input), forward_type)
        predict_RNA = self.RNA_decoder(latent_layer)
        reconstruct_loss = r_loss(predict_RNA, RNA_input)
        kl_div_r = - 0.5 * torch.mean(1 + d - mu.pow(2) - d.exp())
        loss = reconstruct_loss + kl_div_w * kl_div_r
        return loss, reconstruct_loss, kl_div_r
    
    
    def forward_A2A(self, ATAC_input, a_loss, kl_div_w, forward_type):
        latent_layer, mu, d = self.A_translator(self.ATAC_encoder(ATAC_input), forward_type)
        predict_ATAC = self.ATAC_decoder(latent_layer)
        reconstruct_loss = a_loss(predict_ATAC, ATAC_input)
        kl_div_a = - 0.5 * torch.mean(1 + d - mu.pow(2) - d.exp())
        loss = reconstruct_loss + kl_div_w * kl_div_a
        return loss, reconstruct_loss, kl_div_a
        
       
    def forward_translator(self, batch_samples, RNA_input_dim, ATAC_input_dim, a_loss, r_loss, loss_weight, forward_type, kl_div_mean=False):
    
        RNA_input, ATAC_input = torch.split(batch_samples, [RNA_input_dim, ATAC_input_dim], dim=1)
    
        # forward generator
        
        R2 = self.RNA_encoder(RNA_input)
        A2 = self.ATAC_encoder(ATAC_input)
        if forward_type == 'train':
            R2R, R2A, mu_r, sigma_r = self.translator.train_model(R2, 'RNA')
            A2R, A2A, mu_a, sigma_a = self.translator.train_model(A2, 'ATAC')
        elif forward_type == 'test':
            R2R, R2A, mu_r, sigma_r = self.translator.test_model(R2, 'RNA')
            A2R, A2A, mu_a, sigma_a = self.translator.test_model(A2, 'ATAC')        
        
        R2R = self.RNA_decoder(R2R)
        R2A = self.ATAC_decoder(R2A)
        A2R = self.RNA_decoder(A2R)
        A2A = self.ATAC_decoder(A2A)
        
        # reconstruct loss
        lossR2R = r_loss(R2R, RNA_input)
        lossA2R = r_loss(A2R, RNA_input)
        lossR2A = a_loss(R2A, ATAC_input)
        lossA2A = a_loss(A2A, ATAC_input)
        
        # kl divergence
        if kl_div_mean:
            kl_div_r = - 0.5 * torch.mean(1 + sigma_r - mu_r.pow(2) - sigma_r.exp())
            kl_div_a = - 0.5 * torch.mean(1 + sigma_a - mu_a.pow(2) - sigma_a.exp())
        else:
            kl_div_r = torch.clamp(- 0.5 * torch.sum(1 + sigma_r - mu_r.pow(2) - sigma_r.exp()), 0, 10000)
            kl_div_a = torch.clamp(- 0.5 * torch.sum(1 + sigma_a - mu_a.pow(2) - sigma_a.exp()), 0, 10000)
        
        # calculate the loss
        r_loss_w, a_loss_w, d_loss_w, kl_div_R, kl_div_A, kl_div_w = loss_weight
        reconstruct_loss = r_loss_w * (lossR2R + lossA2R) + a_loss_w * (lossR2A + lossA2A)

        kl_div = kl_div_r + kl_div_a
        
        loss_g = kl_div_w * kl_div + reconstruct_loss

        return reconstruct_loss, kl_div, loss_g
    
    
    def forward_discriminator(self, batch_samples, RNA_input_dim, ATAC_input_dim, d_loss, forward_type):
        
        RNA_input, ATAC_input = torch.split(batch_samples, [RNA_input_dim, ATAC_input_dim], dim=1)

        # forward of generator
        R2 = self.RNA_encoder(RNA_input)
        A2 = self.ATAC_encoder(ATAC_input)
        if forward_type == 'train':
            R2R, R2A, mu_r, sigma_r = self.translator.train_model(R2, 'RNA')
            A2R, A2A, mu_a, sigma_a = self.translator.train_model(A2, 'ATAC')
        elif forward_type == 'test':
            R2R, R2A, mu_r, sigma_r = self.translator.test_model(R2, 'RNA')
            A2R, A2A, mu_a, sigma_a = self.translator.test_model(A2, 'ATAC')

        batch_size = batch_samples.shape[0]

        # 1 menas a real data 0 menas a generated data, here use a soft label
        temp1 = np.random.rand(batch_size)
        temp = [0 for item in temp1]
        for i in range(len(temp1)):
            if temp1[i] > 0.8:
                temp[i] = temp1[i]
            elif temp1[i] <= 0.8 and temp1[i] > 0.5:
                temp[i] = 0.8
            elif temp1[i] <= 0.5 and temp1[i] > 0.2:
                temp[i] = 0.2
            else:
                temp[i] = temp1[i]

        input_data_a = torch.stack([A2[i] if temp[i] > 0.5 else R2A[i] for i in range(batch_size)], dim=0)
        input_data_r = torch.stack([R2[i] if temp[i] > 0.5 else A2R[i] for i in range(batch_size)], dim=0)

        predict_atac = self.discriminator_A(input_data_a)
        predict_rna = self.discriminator_R(input_data_r)

        loss1 = d_loss(predict_atac.reshape(batch_size), torch.tensor(temp).cuda().float())
        loss2 = d_loss(predict_rna.reshape(batch_size), torch.tensor(temp).cuda().float())
        return loss1 + loss2
        
    
    def save_model_dict(self, output_path):
        torch.save(self.RNA_encoder.state_dict(), output_path + '/model/RNA_encoder.pt')
        torch.save(self.ATAC_encoder.state_dict(), output_path + '/model/ATAC_encoder.pt')
        torch.save(self.RNA_decoder.state_dict(), output_path + '/model/RNA_decoder.pt')
        torch.save(self.ATAC_decoder.state_dict(), output_path + '/model/ATAC_decoder.pt')
        torch.save(self.R_translator.state_dict(), output_path + '/model/R_translator.pt')
        torch.save(self.A_translator.state_dict(), output_path + '/model/A_translator.pt')
        torch.save(self.translator.state_dict(), output_path + '/model/translator.pt')
        torch.save(self.discriminator_A.state_dict(), output_path + '/model/discriminator_A.pt')
        torch.save(self.discriminator_R.state_dict(), output_path + '/model/discriminator_R.pt')
   

    def evaluation(
        self,
        max_evaluate : int,
        RNA_data,
        ATAC_data,
        id_list
    ):
        mean_test_loss = 0
        mean_test_auroc = 0
        mean_test_aupr = 0
        mean_test_pearson = 0
        mean_test_spearman = 0
        sum_auroc = 0
        sum_aupr = 0
        max_enum = 9
        RNA_input_dim = RNA_data.X.toarray().shape[1]
        ATAC_input_dim = ATAC_data.X.toarray().shape[1]
        
        self.test_dataset = RNA_ATAC_dataset(RNA_data, ATAC_data, id_list, id_list)
        
        self.test_dataloader = DataLoader(self.test_dataset, batch_size=1, shuffle=False, num_workers=self.num_workers)
        
        self.set_eval()
        batch_samples_list_A = []
        # to get every_mean
        for epoch1 in range(1):
            with torch.no_grad():
                for idx, batch_samples in enumerate(self.test_dataloader):

                    if torch.cuda.is_available():
                        batch_samples = batch_samples.cuda().to(torch.float32)

                    RNA_input, ATAC_input = torch.split(batch_samples, [RNA_input_dim, ATAC_input_dim], dim=1)
                    for i in range(len(ATAC_input)):
                        batch_samples_list_A.append(torch.reshape(ATAC_input[i].cpu(),(1,len(ATAC_input[0]))))
            batch_samples_list_A = torch.cat(batch_samples_list_A, dim = 0)
            every_mean = torch.mean(batch_samples_list_A, dim = 0)
        
        with torch.no_grad():
            for idx, batch_samples in enumerate(self.test_dataloader):
                if torch.cuda.is_available():
                    batch_samples = batch_samples.cuda().to(torch.float32)


                RNA_input, ATAC_input = torch.split(batch_samples, [RNA_input_dim, ATAC_input_dim], dim=1)
                R2 = self.RNA_encoder(RNA_input)
                A2 = self.ATAC_encoder(ATAC_input)
                R2R, R2A, mu_r, sigma_r = self.translator.test_model(R2, 'RNA')
                A2R, A2A, mu_a, sigma_a = self.translator.test_model(A2, 'ATAC')
                R2A = self.ATAC_decoder(R2A)
                A2R = self.RNA_decoder(A2R)

                atac_input = ATAC_input.cpu().numpy()
                rna_input = RNA_input.cpu().numpy()
                r2a = R2A.cpu().detach().numpy()
                a2r = A2R.cpu().detach().numpy()
                ss = sum(atac_input)

                for i in range(len(atac_input)):
                    a=AUROC(r2a[i,:],atac_input[i,:], every_mean)
                    if a != 0:
                        mean_test_auroc += a
                        sum_auroc += 1
                    else:
                        mean_test_auroc += a
                    b = AUPR_norm(r2a[i,:],atac_input[i,:], every_mean)
                    if b!= 0:
                        mean_test_aupr += b
                        sum_aupr += 1
                    else:
                        mean_test_aupr += b

                    mean_test_pearson += pearson(a2r[i,:],rna_input[i,:])
                    mean_test_spearman += spearman(a2r[i,:],rna_input[i,:])
                print(idx)
                        
                if idx == max_evaluate-1:
                    break

        mean_test_auroc /= sum_auroc
        mean_test_aupr /= sum_aupr
        mean_test_pearson /= max_evaluate
        mean_test_spearman /= max_evaluate
 
        return [mean_test_auroc,mean_test_aupr,mean_test_pearson,mean_test_spearman]



    def train(
        self,
        loss_weight: list,
        train_id_r: list,
        train_id_a: list,
        validation_id_r: list,
        validation_id_a: list,
        R_encoder_lr: float = 0.001,
        A_encoder_lr: float = 0.001,
        R_decoder_lr: float = 0.001,
        A_decoder_lr: float = 0.001,
        R_translator_lr: float = 0.001,
        A_translator_lr: float = 0.001,
        translator_lr: float = 0.001,
        discriminator_lr: float = 0.005,
        R2R_pretrain_epoch: int = 100,
        A2A_pretrain_epoch: int = 100,
        lock_encoder_and_decoder: bool = False,
        translator_epoch: int = 200,
        patience: int = 50,
        batch_size: int = 64,
        r_loss = nn.MSELoss(size_average=True),
        a_loss = nn.MSELoss(size_average=True),
        d_loss = nn.BCELoss(size_average=True),
        output_path: str = None,
        seed: int = 19193,
        kl_mean: bool = True,
        R_pretrain_kl_warmup: int = 50,
        A_pretrain_kl_warmup: int = 50,
        translation_kl_warmup: int = 50,
        load_model: str = None,
        logging_path: str = None
    ):
        """
        Training for model. Some parameters need information about data, please see in Tutorial.
        
        Parameters
        ----------
        loss_weight: list
            list of loss weight for [r_loss, a_loss, d_loss, kl_div_R, kl_div_A, kl_div_all].
        
        train_id_r: list
            list of RNA data cell ids for training.
            
        train_id_a: list
            list of ADT data cell ids for training.
            
        validation_id_r: list
            list of RNA data cell ids for validation.
        
        validation_id_a: list
            list of ADT data cell ids for validation.

        R_encoder_lr: float
            learning rate of RNA encoder, default 0.001.
            
        A_encoder_lr: float
            learning rate of ADT encoder, default 0.001.
            
        R_decoder_lr: float
            learning rate of RNA decoder, default 0.001.
            
        A_decoder_lr: float
            learning rate of ADT decoder, default 0.001.
       
        R_translator_lr: float
            learning rate of RNA pretrain translator, default 0.001.
            
        A_translator_lr: float
            learning rate of ADT pretrain translator, default 0.001.
            
        translator_lr: float
            learning rate of translator, default 0.001.
            
        discriminator_lr: float
            learning rate of discriminator, default 0.005.
            
        R2R_pretrain_epoch: int
            max epoch for pretrain RNA autoencoder, default 100.
            
        A2A_pretrain_epoch: int
            max epoch for pretrain ADT autoencoder, default 100.
            
        lock_encoder_and_decoder: bool
            lock the pretrained encoder and decoder or not, default False.
            
        translator_epoch: int
            max epoch for train translator, default 200.
            
        patience: int
            patience for loss on validation, default 50.
            
        batch_size: int
            batch size for training and validation, default 64.
            
        r_loss
            loss function for RNA reconstruction, default nn.MSELoss(size_average=True).
            
        a_loss
            loss function for ADT reconstruction, default nn.MSELoss(size_average=True).
            
        d_loss
            loss function for discriminator, default nn.BCELoss(size_average=True).
            
        output_path: str
            file path for model output, default None.
            
        seed: int
            set up the random seed, default 19193.
            
        kl_mean: bool
            size average for kl divergence or not, default True.
            
        R_pretrain_kl_warmup: int
            epoch of linear weight warm up for kl divergence in RNA pretrain, default 50.
        
        A_pretrain_kl_warmup: int
            epoch of linear weight warm up for kl divergence in ADT pretrain, default 50.
        
        translation_kl_warmup: int
            epoch of linear weight warm up for kl divergence in translator pretrain, default 50.
            
        load_model: str
            the path for loading model if needed, else set it None, default None.
            
        logging_path: str
            the path for output process logging, if not save, set it None, default None.
            
        """
        if not logging_path is None:
            file_handle=open(logging_path + '/Parameters_Record.txt',mode='a')
            file_handle.writelines([
                '------------------------------\n'
                'Train Parameters\n'
                'R_encoder_lr: '+str(R_encoder_lr)+'\n',
                'A_encoder_lr: '+str(A_encoder_lr)+'\n',
                'R_decoder_lr: '+str(R_decoder_lr)+'\n',
                'A_decoder_lr: '+str(A_decoder_lr)+'\n',
                'R_translator_lr: '+str(R_translator_lr)+'\n',
                'A_translator_lr: '+str(A_translator_lr)+'\n',
                'translator_lr: '+str(translator_lr)+'\n',
                'discriminator_lr: '+str(discriminator_lr)+'\n',
                'R2R_pretrain_epoch: '+str(R2R_pretrain_epoch)+'\n',
                'A2A_pretrain_epoch: '+str(A2A_pretrain_epoch)+'\n',
                'lock_encoder_and_decoder: '+str(lock_encoder_and_decoder)+'\n',
                'translator_epoch: '+str(translator_epoch)+'\n',
                'patience: '+str(patience)+'\n',
                'batch_size: '+str(batch_size)+'\n',
                'r_loss: '+str(r_loss)+'\n',
                'a_loss: '+str(a_loss)+'\n',
                'd_loss: '+str(d_loss)+'\n',
                'loss_weight: '+str(loss_weight)+'\n',
                'seed: '+str(seed)+'\n',
                'kl_mean: '+str(kl_mean)+'\n',
                'R_pretrain_kl_warmup: '+str(R_pretrain_kl_warmup)+'\n',
                'A_pretrain_kl_warmup: '+str(A_pretrain_kl_warmup)+'\n',
                'translation_kl_warmup: '+str(translation_kl_warmup)+'\n',
                'load_model: '+str(load_model)+'\n'
            ])
            file_handle.close()
            
        my_logger = create_logger(name='Trainer', ch=True, fh=False, levelname=logging.INFO, overwrite=False)
        
        self.is_train_finished = False
        
        if output_path is None:
            output_path = '.'
        
        if not load_model is None:
            my_logger.info('load pretrained model from path: '+str(load_model)+'/model/')
            self.RNA_encoder.load_state_dict(torch.load(load_model + '/model/RNA_encoder.pt'))
            self.ATAC_encoder.load_state_dict(torch.load(load_model + '/model/ATAC_encoder.pt'))
            self.RNA_decoder.load_state_dict(torch.load(load_model + '/model/RNA_decoder.pt'))
            self.ATAC_decoder.load_state_dict(torch.load(load_model + '/model/ATAC_decoder.pt'))
            self.translator.load_state_dict(torch.load(load_model + '/model/translator.pt'))
            self.discriminator_A.load_state_dict(torch.load(load_model + '/model/discriminator_A.pt'))
            self.discriminator_R.load_state_dict(torch.load(load_model + '/model/discriminator_R.pt'))
            
        if not seed is None:
            setup_seed(seed)
        
        RNA_input_dim = self.RNA_data.shape[1]
        ATAC_input_dim = self.ATAC_data.shape[1]
        cell_count = len(train_id_r)
        
        # define the dataset and dataloader for train and validation
        self.train_dataset = RNA_ATAC_dataset(self.RNA_data, self.ATAC_data, train_id_r, train_id_a)
        self.validation_dataset = RNA_ATAC_dataset(self.RNA_data, self.ATAC_data, validation_id_r, validation_id_a)

        if cell_count % batch_size == 1:
            self.train_dataloader = DataLoader(self.train_dataset, batch_size=batch_size, shuffle=True, num_workers=self.num_workers, drop_last=True)
        else:
            self.train_dataloader = DataLoader(self.train_dataset, batch_size=batch_size, shuffle=True, num_workers=self.num_workers)
        self.validation_dataloader = DataLoader(self.validation_dataset, batch_size=batch_size, shuffle=False, num_workers=self.num_workers)
        
        self.optimizer_R_encoder = torch.optim.Adam(self.RNA_encoder.parameters(), lr=R_encoder_lr)
        self.optimizer_A_encoder = torch.optim.Adam(self.ATAC_encoder.parameters(), lr=A_encoder_lr, weight_decay=0)
        self.optimizer_R_decoder = torch.optim.Adam(self.RNA_decoder.parameters(), lr=R_decoder_lr)
        self.optimizer_A_decoder = torch.optim.Adam(self.ATAC_decoder.parameters(), lr=A_decoder_lr, weight_decay=0)
        self.optimizer_R_translator = torch.optim.Adam(self.R_translator.parameters(), lr=R_translator_lr)
        self.optimizer_A_translator = torch.optim.Adam(self.A_translator.parameters(), lr=A_translator_lr)
        self.optimizer_translator = torch.optim.Adam(self.translator.parameters(), lr=translator_lr)
        self.optimizer_discriminator_A = torch.optim.SGD(self.discriminator_A.parameters(), lr=discriminator_lr)
        self.optimizer_discriminator_R = torch.optim.SGD(self.discriminator_R.parameters(), lr=discriminator_lr)
        
        """ eraly stop for model """
        self.early_stopping_R2R = EarlyStopping(patience=patience, verbose=False)
        self.early_stopping_A2A = EarlyStopping(patience=patience, verbose=False)
        self.early_stopping_all = EarlyStopping(patience=patience, verbose=False)
        
        if not os.path.exists(output_path + '/model'):
            os.mkdir(output_path + '/model')
        
        """ pretrain for RNA and ADT """
        my_logger.info('RNA pretraining ...')
        pretrain_r_loss, pretrain_r_kl, pretrain_r_loss_val, pretrain_r_kl_val = [], [], [], []
        with tqdm(total = R2R_pretrain_epoch, ncols=100) as pbar:
            pbar.set_description('RNA pretrain')
            for epoch in range(R2R_pretrain_epoch):
                pretrain_r_loss_, pretrain_r_kl_, pretrain_r_loss_val_, pretrain_r_kl_val_ = [], [], [], []
                self.set_train()
                for idx, batch_samples in enumerate(self.train_dataloader): 

                    if torch.cuda.is_available():
                        batch_samples = batch_samples.cuda().to(torch.float32)

                    RNA_input, ATAC_input = torch.split(batch_samples, [RNA_input_dim, ATAC_input_dim], dim=1)

                    """ pretrain for RNA """
                    weight_temp = loss_weight.copy()
                    if epoch < R_pretrain_kl_warmup:
                        weight_temp[3] = loss_weight[3] * epoch / R_pretrain_kl_warmup

                    loss, reconstruct_loss, kl_div_r = self.forward_R2R(RNA_input, r_loss, weight_temp[3], 'train')
                    self.optimizer_R_encoder.zero_grad()
                    self.optimizer_R_decoder.zero_grad()
                    self.optimizer_R_translator.zero_grad()
                    loss.backward()
                    self.optimizer_R_encoder.step()
                    self.optimizer_R_decoder.step()
                    self.optimizer_R_translator.step()

                    pretrain_r_loss_.append(reconstruct_loss.item())
                    pretrain_r_kl_.append(kl_div_r.item())

                self.set_eval()
                for idx, batch_samples in enumerate(self.validation_dataloader):

                    if torch.cuda.is_available():
                        batch_samples = batch_samples.cuda().to(torch.float32)

                    RNA_input, ATAC_input = torch.split(batch_samples, [RNA_input_dim, ATAC_input_dim], dim=1)

                    loss, reconstruct_loss, kl_div_r = self.forward_R2R(RNA_input, r_loss, weight_temp[3], 'test')

                    pretrain_r_loss_val_.append(reconstruct_loss.item())
                    pretrain_r_kl_val_.append(kl_div_r.item())

                pretrain_r_loss.append(np.mean(pretrain_r_loss_))
                pretrain_r_kl.append(np.mean(pretrain_r_kl_))
                pretrain_r_loss_val.append(np.mean(pretrain_r_loss_val_))
                pretrain_r_kl_val.append(np.mean(pretrain_r_kl_val_))

                self.early_stopping_R2R(np.mean(pretrain_r_loss_val_), self, output_path)
                time.sleep(0.01)
                pbar.update(1)
                pbar.set_postfix(
                    train='{:.4f}'.format(np.mean(pretrain_r_loss_val_)), 
                    val='{:.4f}'.format(np.mean(pretrain_r_loss_)))
                
                if self.early_stopping_R2R.early_stop:
                    my_logger.info('RNA pretraining early stop, validation loss does not improve in '+str(patience)+' epoches!')
                    self.RNA_encoder.load_state_dict(torch.load(output_path + '/model/RNA_encoder.pt'))
                    self.RNA_decoder.load_state_dict(torch.load(output_path + '/model/RNA_decoder.pt'))
                    self.R_translator.load_state_dict(torch.load(output_path + '/model/R_translator.pt'))
                    break
        
        
        pretrain_a_loss, pretrain_a_kl, pretrain_a_loss_val, pretrain_a_kl_val = [], [], [], []
        my_logger.info('ADT pretraining ...')
        with tqdm(total = A2A_pretrain_epoch, ncols=100) as pbar:
            pbar.set_description('ADT pretrain')
            for epoch in range(A2A_pretrain_epoch):
                pretrain_a_loss_, pretrain_a_kl_, pretrain_a_loss_val_, pretrain_a_kl_val_ = [], [], [], []
                self.set_train()
                for idx, batch_samples in enumerate(self.train_dataloader): 

                    if torch.cuda.is_available():
                        batch_samples = batch_samples.cuda().to(torch.float32)

                    RNA_input, ATAC_input = torch.split(batch_samples, [RNA_input_dim, ATAC_input_dim], dim=1)

                    """ pretrain for ADT """
                    weight_temp = loss_weight.copy()
                    if epoch < A_pretrain_kl_warmup:
                        weight_temp[4] = loss_weight[4] * epoch / A_pretrain_kl_warmup

                    loss, reconstruct_loss, kl_div_a = self.forward_A2A(ATAC_input, a_loss, weight_temp[4], 'train')
                    self.optimizer_A_encoder.zero_grad()
                    self.optimizer_A_decoder.zero_grad()
                    self.optimizer_A_translator.zero_grad()
                    loss.backward()
                    self.optimizer_A_encoder.step()
                    self.optimizer_A_decoder.step()
                    self.optimizer_A_translator.step()

                    pretrain_a_loss_.append(reconstruct_loss.item())
                    pretrain_a_kl_.append(kl_div_a.item())

                self.set_eval()
                for idx, batch_samples in enumerate(self.validation_dataloader):

                    if torch.cuda.is_available():
                        batch_samples = batch_samples.cuda().to(torch.float32)

                    RNA_input, ATAC_input = torch.split(batch_samples, [RNA_input_dim, ATAC_input_dim], dim=1)

                    loss, reconstruct_loss, kl_div_a = self.forward_A2A(ATAC_input, a_loss, weight_temp[4], 'test')

                    pretrain_a_loss_val_.append(reconstruct_loss.item())
                    pretrain_a_kl_val_.append(kl_div_a.item())

                pretrain_a_loss.append(np.mean(pretrain_a_loss_))
                pretrain_a_kl.append(np.mean(pretrain_a_kl_))
                pretrain_a_loss_val.append(np.mean(pretrain_a_loss_val_))
                pretrain_a_kl_val.append(np.mean(pretrain_a_kl_val_))

                self.early_stopping_A2A(np.mean(pretrain_a_loss_val_), self, output_path)
                time.sleep(0.01)
                pbar.update(1)
                pbar.set_postfix(
                    train='{:.4f}'.format(np.mean(pretrain_a_loss_val_)), 
                    val='{:.4f}'.format(np.mean(pretrain_a_loss_)))
                
                if self.early_stopping_A2A.early_stop:
                    my_logger.info('ADT pretraining early stop, validation loss does not improve in '+str(patience)+' epoches!')
                    self.ATAC_encoder.load_state_dict(torch.load(output_path + '/model/ATAC_encoder.pt'))
                    self.ATAC_decoder.load_state_dict(torch.load(output_path + '/model/ATAC_decoder.pt'))
                    self.A_translator.load_state_dict(torch.load(output_path + '/model/A_translator.pt'))
                    break

        
        """ train for translator and discriminator """
        train_loss, train_kl, train_discriminator, train_loss_val, train_kl_val, train_discriminator_val = [], [], [], [], [], []
        my_logger.info('Combine training ...')
        with tqdm(total = translator_epoch, ncols=100) as pbar:
            pbar.set_description('Combine training')
            for epoch in range(translator_epoch):
                train_loss_, train_kl_, train_discriminator_, train_loss_val_, train_kl_val_, train_discriminator_val_ = [], [], [], [], [], []
                self.set_train()
                for idx, batch_samples in enumerate(self.train_dataloader):

                    if torch.cuda.is_available():
                        batch_samples = batch_samples.cuda().to(torch.float32)

                    RNA_input, ATAC_input = torch.split(batch_samples, [RNA_input_dim, ATAC_input_dim], dim=1)

                    """ train for discriminator """
                    loss_d = self.forward_discriminator(batch_samples, RNA_input_dim, ATAC_input_dim, d_loss, 'train')
                    self.optimizer_discriminator_R.zero_grad()
                    self.optimizer_discriminator_A.zero_grad()
                    loss_d.backward()
                    self.optimizer_discriminator_R.step()
                    self.optimizer_discriminator_A.step()

                    """ train for generator """
                    weight_temp = loss_weight.copy()
                    if epoch < translation_kl_warmup:
                        weight_temp[5] = loss_weight[5] * epoch / translation_kl_warmup
                    loss_d = self.forward_discriminator(batch_samples, RNA_input_dim, ATAC_input_dim, d_loss, 'train')
                    reconstruct_loss, kl_div, loss_g = self.forward_translator(batch_samples, RNA_input_dim, ATAC_input_dim, a_loss, r_loss, weight_temp, 'train', kl_mean)

                    if loss_d.item() < 1.35:
                        loss_g -= loss_weight[2] * loss_d

                    self.optimizer_translator.zero_grad()
                    if not lock_encoder_and_decoder:
                        self.optimizer_R_encoder.zero_grad()
                        self.optimizer_A_encoder.zero_grad()
                        self.optimizer_R_decoder.zero_grad()
                        self.optimizer_A_decoder.zero_grad()
                    loss_g.backward()
                    self.optimizer_translator.step()
                    if not lock_encoder_and_decoder:
                        self.optimizer_R_encoder.step()
                        self.optimizer_A_encoder.step()
                        self.optimizer_R_decoder.step()
                        self.optimizer_A_decoder.step()

                    train_loss_.append(reconstruct_loss.item())
                    train_kl_.append(kl_div.item()) 
                    train_discriminator_.append(loss_d.item())

                self.set_eval()
                for idx, batch_samples in enumerate(self.validation_dataloader):

                    if torch.cuda.is_available():
                        batch_samples = batch_samples.cuda().to(torch.float32)                    

                    RNA_input, ATAC_input = torch.split(batch_samples, [RNA_input_dim, ATAC_input_dim], dim=1)

                    """ test for discriminator """
                    loss_d = self.forward_discriminator(batch_samples, RNA_input_dim, ATAC_input_dim, d_loss, 'test')

                    """ test for generator """
                    loss_d = self.forward_discriminator(batch_samples, RNA_input_dim, ATAC_input_dim, d_loss, 'test')
                    reconstruct_loss, kl_div, loss_g = self.forward_translator(batch_samples, RNA_input_dim, ATAC_input_dim, a_loss, r_loss, weight_temp, 'train', kl_mean)
                    loss_g -= loss_weight[2] * loss_d

                    train_loss_val_.append(reconstruct_loss.item())
                    train_kl_val_.append(kl_div.item()) 
                    train_discriminator_val_.append(loss_d.item())

                train_loss.append(np.mean(train_loss_))
                train_kl.append(np.mean(train_kl_))
                train_discriminator.append(np.mean(train_discriminator_))
                train_loss_val.append(np.mean(train_loss_val_))
                train_kl_val.append(np.mean(train_kl_val_))
                train_discriminator_val.append(np.mean(train_discriminator_val_))
                self.early_stopping_all(np.mean(train_loss_val_), self, output_path)
                
                time.sleep(0.01)
                pbar.update(1)
                pbar.set_postfix(
                    train='{:.4f}'.format(np.mean(train_loss_val_)), 
                    val='{:.4f}'.format(np.mean(train_loss_)))
                
                if self.early_stopping_all.early_stop:
                    my_logger.info('Combine training early stop, validation loss does not improve in '+str(patience)+' epoches!')
                    self.RNA_encoder.load_state_dict(torch.load(output_path + '/model/RNA_encoder.pt'))
                    self.ATAC_encoder.load_state_dict(torch.load(output_path + '/model/ATAC_encoder.pt'))
                    self.RNA_decoder.load_state_dict(torch.load(output_path + '/model/RNA_decoder.pt'))
                    self.ATAC_decoder.load_state_dict(torch.load(output_path + '/model/ATAC_decoder.pt'))
                    self.translator.load_state_dict(torch.load(output_path + '/model/translator.pt'))
                    self.discriminator_A.load_state_dict(torch.load(output_path + '/model/discriminator_A.pt'))
                    self.discriminator_R.load_state_dict(torch.load(output_path + '/model/discriminator_R.pt'))
                    break
        
        self.save_model_dict(output_path)
            
        self.is_train_finished = True
        
        record_loss_log(
            pretrain_r_loss,
            pretrain_r_kl,
            pretrain_r_loss_val,
            pretrain_r_kl_val,
            pretrain_a_loss,
            pretrain_a_kl,
            pretrain_a_loss_val,
            pretrain_a_kl_val,
            train_loss,
            train_kl,
            train_discriminator,
            train_loss_val,
            train_kl_val,
            train_discriminator_val,
            output_path
        )

    def test(
        self,
        test_id_r: list,
        test_id_a: list,
        model_path: str = None,
        load_model: bool = True,
        output_path: str = None,
        test_evaluate: bool = True,
        test_cluster: bool = True,
        test_figure: bool = True,
        output_data: bool = False,
        return_predict: bool = False
    ):
        """
        Test for model.
        
        Parameters
        ----------            
        train_id_r: list
            list of RNA data cell ids for training.
            
        train_id_a: list
            list of ATAC data cell ids for training.
            
        model_path: str
            path for load trained model, default None.
            
        load_model: bool
            load the pretrained model or not, deafult True.
            
        output_path: str
            file path for model output, default None.
            
        test_evaludate: bool
            test correlation or not, deafult True.
            
        test_cluster: bool
            test clustrer index or not, deafult True.
            
        test_figure: bool
            test tSNE or not, deafult True.
            
        output_data: bool
            output the predicted test data to file or not, deafult False.
            
        return_predict: bool
            return predict or not, if True, output (A2R_predict, R2A_predict) as returns, deafult False.
            
        """
        """ load model from model_path if need """
        my_logger = create_logger(name='Tester', ch=True, fh=False, levelname=logging.INFO, overwrite=False)
        
        if output_path is None:
            output_path = '.'
        
        if load_model:
            my_logger.info('load trained model from path: '+str(model_path)+'/model')
            self.RNA_encoder.load_state_dict(torch.load(model_path + '/model/RNA_encoder.pt'))
            self.ATAC_encoder.load_state_dict(torch.load(model_path + '/model/ATAC_encoder.pt'))
            self.RNA_decoder.load_state_dict(torch.load(model_path + '/model/RNA_decoder.pt'))
            self.ATAC_decoder.load_state_dict(torch.load(model_path + '/model/ATAC_decoder.pt'))
            self.translator.load_state_dict(torch.load(model_path + '/model/translator.pt'))
        
        """ load data """
        RNA_input_dim = self.RNA_data.shape[1]
        ATAC_input_dim = self.ATAC_data.shape[1]
        
        self.R_test_dataset = Single_omics_dataset(self.RNA_data, test_id_r)
        self.A_test_dataset = Single_omics_dataset(self.ATAC_data, test_id_a)
        self.R_test_dataloader = DataLoader(self.R_test_dataset, batch_size=64, shuffle=False, num_workers=self.num_workers)
        self.A_test_dataloader = DataLoader(self.A_test_dataset, batch_size=64, shuffle=False, num_workers=self.num_workers)

        self.set_eval()
        my_logger.info('get predicting ...')
        """ record the predicted data """
        R2A_predict = []
        A2R_predict = []
        R2_predict = []
        A2_predict = []
        with torch.no_grad():
            with tqdm(total = len(self.R_test_dataloader), ncols=100) as pbar:
                pbar.set_description('RNA to ADT predicting...')
                for idx, batch_samples in enumerate(self.R_test_dataloader):
                    if torch.cuda.is_available():
                        batch_samples = batch_samples.cuda().to(torch.float32)

                    R2 = self.RNA_encoder(batch_samples)
                    R2R, R2A, mu_r, sigma_r = self.translator.test_model(R2, 'RNA')
                    R2A = self.ATAC_decoder(R2A)
                    
                    R2A_predict.append(R2A.cpu())
                    R2_predict.append(mu_r.cpu())
                    
                    time.sleep(0.01)
                    pbar.update(1)

        with torch.no_grad():
            with tqdm(total = len(self.A_test_dataloader), ncols=100) as pbar:
                pbar.set_description('ADT to RNA predicting...')
                for idx, batch_samples in enumerate(self.A_test_dataloader):
                    if torch.cuda.is_available():
                        batch_samples = batch_samples.cuda().to(torch.float32)

                    A2 = self.ATAC_encoder(batch_samples)
                    A2R, A2A, mu_a, sigma_a = self.translator.test_model(A2, 'ATAC')
                    A2R = self.RNA_decoder(A2R)

                    A2R_predict.append(A2R.cpu())                       
                    A2_predict.append(mu_a.cpu())
                    
                    time.sleep(0.01)
                    pbar.update(1)
        
        R2A_predict = tensor2adata(R2A_predict)
        A2R_predict = tensor2adata(A2R_predict)
        
        A2R_predict.obs = self.ATAC_data_obs.iloc[test_id_a, :]
        R2A_predict.obs = self.RNA_data_obs.iloc[test_id_r, :]

        """ draw tsne if needed """
        if test_figure:
            my_logger.info('drawing tsne figures ...')
            fig_A2R = draw_tsne(A2R_predict, 'a2r', 'cell_type')
            fig_R2A = draw_tsne(R2A_predict, 'r2a', 'cell_type')
            fig_list = [fig_A2R, fig_R2A]
            with PdfPages(output_path + '/tSNE.pdf') as pdf:
                for i in range(len(fig_list)):
                    pdf.savefig(figure=fig_list[i], dpi=200, bbox_inches='tight')
                    plt.close()
        else:
            my_logger.info('calculate neighbors graph for following test ...')
            sc.pp.pca(A2R_predict)
            sc.pp.neighbors(A2R_predict)
            sc.pp.pca(R2A_predict)
            sc.pp.neighbors(R2A_predict)
                    
        """ test cluster index if needed """
        if test_cluster:
            index_R2A = calculate_cluster_index(R2A_predict)
            index_A2R = calculate_cluster_index(A2R_predict)
            
            index_matrix = pd.DataFrame([index_R2A, index_A2R])
            index_matrix.columns = ['ARI', 'AMI', 'NMI', 'HOM', 'COM']
            index_matrix.index = ['R2A', 'A2R']
            index_matrix.to_csv(output_path + '/cluster_index.csv')
            
        if test_evaluate:
            my_logger.info('evaluating correlation ...')
            pearson_list = []
            spearman_list = []
            R2A_predict_dense = R2A_predict.X.todense()
            ADT_data_train_dense = self.ATAC_data[test_id_a, :]
            for i in range(R2A_predict_dense.shape[0]):
                pearson_list.append(pearson(R2A_predict_dense[i].A[0], ADT_data_train_dense[i]))
                spearman_list.append(spearman(R2A_predict_dense[i].A[0], ADT_data_train_dense[i]))
            pearson_index = np.mean(pearson_list)
            spearma_index = np.mean(spearman_list)
            index_matrix = pd.DataFrame([[pearson_index, spearma_index]])
            index_matrix.columns = ['pearson', 'spearman']
            index_matrix.index = ['R2A']
            index_matrix.to_csv(output_path + '/relation.csv')

        """ save predicted model if needed """
        if output_data and not os.path.exists(output_path + '/predict'):
            my_logger.warning('trying to write predict to path: '+str(output_path)+'/predict')
            os.mkdir(output_path + '/predict')
            A2R_predict.write_h5ad(output_path + '/predict/A2R.h5ad')
            R2A_predict.write_h5ad(output_path + '/predict/R2A.h5ad')
            
        if return_predict:
            return A2R_predict, R2A_predict


import scanpy as sc
import anndata as ad
from scipy import sparse
import pandas as pd


def draw_tsne(data, title, color):
    """
    Visualization of prediction using tSNE Algorithm.
    
    Parameters
    ----------
    data: Anndata
        Anndata for visualization.
        
    title: str
        title of picture.

    color: list
        color of picture

    Returns
    ----------
    fig
        figure with specific color and title.
    """

    sc.settings.set_figure_params(dpi=120, facecolor='white')
    sc.pp.pca(data)
    sc.pp.neighbors(data)
    sc.tl.tsne(data)
    fig = sc.pl.tsne(data, color=color, title=title, return_fig=True)
    return fig


def draw_reg_plot(eval_adata,
                  cell_type,
                  reg_type='mean',
                  axis_keys={"x": "pred", "y": "stimulated"},
                  condition_key='condition',
                  gene_draw=None,
                  top_gene_list=None,
                  save_path=None,
                  title=None,
                  show=True,
                  fontsize=14
                 ):
    import pandas as pd
    import seaborn as sns
    import numpy as np
    import scanpy as sc
    import matplotlib.pyplot as plt
    from adjustText import adjust_text
    
    df_case = eval_adata[(eval_adata.obs[condition_key]==axis_keys["y"])].to_df()
    df_pred = eval_adata[(eval_adata.obs[condition_key]==axis_keys["x"])].to_df()
    if reg_type=='mean':
        mean_case = df_case.mean().values.reshape(-1, 1)
        mean_pred = df_pred.mean().values.reshape(-1, 1)
    elif reg_type=='var':
        mean_case = df_case.var().values.reshape(-1, 1)
        mean_pred = df_pred.var().values.reshape(-1, 1)
    data = np.hstack((mean_case, mean_pred))
    data_df = pd.DataFrame(data, columns=['case', 'predict'], index=df_case.columns)
    
    sns.set(color_codes=True)
    
    fig = plt.figure()
    ax = fig.add_axes([0,0,1,1])
    
    sns.regplot(x='case', y='predict', data=data_df, ax=ax)
    if gene_draw is not None:
        texts = []
        x = mean_case
        y = mean_pred
        for i in gene_draw:
            j = eval_adata.var_names.tolist().index(i)
            x_bar = x[j]
            y_bar = y[j]
            texts.append(plt.text(x_bar, y_bar, i, fontsize=11, color="black"))
            ax.plot(x_bar, y_bar, "o", color="red", markersize=5)
        adjust_text(
                texts,
                x=x,
                y=y,
                arrowprops=dict(arrowstyle="->", color="grey", lw=0.5),
                force_points=(0.0, 0.0),
        )
    if top_gene_list is not None:
        data_deg = data_df.loc[top_gene_list, :]
        r_top = round(data_deg['case'].corr(data_deg['predict'], method='pearson'), 3)
        xt = 0.1 * np.max(data_df['case'])
        yt = 0.85 * np.max(data_df['predict'])
        ax.text(xt, yt, s='$R^2_{top 100 genes}$=' + str(round(r_top*r_top,3)), fontsize=fontsize, color='black')
    r = round(data_df['case'].corr(data_df['predict'], method='pearson'), 3)
    xt = 0.1 * np.max(data_df['case'])
    yt = 0.75 * np.max(data_df['predict'])
    ax.text(xt, yt, s='$R^2_{all genes}$=' + str(round(r*r,3)), fontsize=fontsize, color='black')
    if title:
        ax.set_title(title)
    else:
        ax.set_title('The Linear Regression of True and Predict Expression '+ reg_type +' of ' + cell_type)

    return round(r*r,3),round(r_top*r_top,3), fig